"""A service for interacting with the filesystem."""

import os
from typing import List

import structlog

LOGGER = structlog.get_logger(__name__)


class FileService:
    """A service for interacting with the filesystem."""

    @staticmethod
    def write_windows_install_paths(target_file: str, paths: List[str]) -> None:
        """
        Write the given list of paths to the target file.

        :param target_file: File to write paths to.
        :param paths: Paths to write to file.
        """
        with open(target_file, "a") as out:
            if os.stat(target_file).st_size > 0:
                out.write(os.pathsep)
            out.write(os.pathsep.join(paths))

        LOGGER.info(f"Finished writing binary paths on Windows to {target_file}")

    @staticmethod
    def append_lines_to_file(file_name: str, content: List[str]) -> None:
        """
        Append the given content to the specified file.

        :param file_name: File to write to.
        :param content: Content to write to file.
        """
        with open(file_name, "a") as out:
            out.write("\n".join(content))

    @staticmethod
    def delete_file(file_name: str) -> None:
        """Delete the given file from the filesystem."""
        os.remove(file_name)



"""A service for working with evergreen."""
from __future__ import annotations

from typing import Dict, Iterator, List, NamedTuple, Optional

import inject
import structlog
from evergreen import EvergreenApi, Task, Version
from requests.exceptions import HTTPError

from db_contrib_tool.config import SetupReproEnvConfig, Tasks

MONGO_PROJECT_PREFIX = "mongodb-mongo-"

LOGGER = structlog.get_logger(__name__)


class ArtifactUrls(NamedTuple):
    """
    Collection of URL containing build artifacts associated with a build.

    * urls: Dictionary of named artifacts to artifact URLs.
    * project_identifier: ID of evergreen project artifacts belong to.
    """

    urls: Dict[str, str]
    project_identifier: str


class RequiredTasks(NamedTuple):
    """Tasks relevant for multiversion setup."""

    symbols_task: Task
    binary_task: Task
    push_task: Optional[Task]

    @classmethod
    def from_task_list(
        cls, task_list: List[Task], targeted_tasks: Tasks, require_push: bool
    ) -> Optional[RequiredTasks]:
        """
        Collect the required tasks from the given task list.

        :param task_list: List of all tasks.
        :param targeted_tasks: Configuration of which tasks to collect.
        :param require_push: If True, require the push task be found.
        :return: Collection of tasks that fulfill the required task criteria.
        """
        binary_task = None
        symbols_task = None
        push_task = None

        for evg_task in task_list:
            if binary_task is None and targeted_tasks.is_task_binary(evg_task):
                binary_task = evg_task

            if push_task is None and targeted_tasks.is_task_push(evg_task):
                push_task = evg_task

            if symbols_task is None and targeted_tasks.is_task_symbols(evg_task):
                symbols_task = evg_task

            if binary_task and symbols_task:
                if require_push and push_task is None:
                    continue

                return cls(
                    binary_task=binary_task,
                    symbols_task=symbols_task,
                    push_task=push_task,
                )

        LOGGER.debug(
            "Missing required tasks",
            binary_task=binary_task,
            symbols_task=symbols_task,
            push_task=push_task,
            require_push=require_push,
        )

        return None

    def get_artifact_urls(self) -> Optional[ArtifactUrls]:
        """
        Get a collection of the relevant artifact URLs form the targeted tasks.

        :return: Artifact URLs for targeted tasks.
        """
        if not self.binary_task.is_success():
            LOGGER.warning(
                "binary task was not successful",
                task_id=self.binary_task.task_id,
                status=self.binary_task.status,
            )
            return None

        if self.push_task is not None and not self.push_task.is_success():
            LOGGER.warning(
                "push task was not successful",
                task_id=self.push_task.task_id,
                status=self.push_task.status,
            )
            return None

        LOGGER.info(
            "Required evergreen task(s) were successful.",
            push_task_required=self.push_task is not None,
            binary_task_id=self.binary_task.task_id,
        )

        artifact_urls = {artifact.name: artifact.url for artifact in self.binary_task.artifacts}

        if self.symbols_task.is_success():
            for artifact in self.symbols_task.artifacts:
                artifact_urls[artifact.name] = artifact.url
        else:
            LOGGER.warning(
                "debug symbol task was unsuccessful",
                archive_symbols_task=self.symbols_task.task_id,
            )

        return ArtifactUrls(
            urls=artifact_urls, project_identifier=self.binary_task.project_identifier
        )


class EvergreenService:
    """A service for working with evergreen."""

    @inject.autoparams()
    def __init__(self, evg_api: EvergreenApi, setup_repro_config: SetupReproEnvConfig) -> None:
        """
        Initialize the service.

        :param evg_api: Evergreen API client.
        """
        self.evg_api = evg_api
        self.setup_repro_config = setup_repro_config

    def get_task(self, task_id: str) -> Task:
        """Get evergreen task by task id."""
        return self.evg_api.task_by_id(task_id)

    def query_task_existence(self, task_id: str) -> bool:
        """
        Query if the given evergreen task exists.

        :param task_id: ID of evergreen task to query.
        :returns: True if the task exists.
        """
        try:
            task = self.evg_api.task_by_id(task_id)
            return task is not None
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            LOGGER.warning("Unexpected error returned by Evergreen.", exc_info=True)
            raise err

    def get_version(self, version_id: str) -> Version:
        """
        Get the requested evergreen version.

        :param version_id: ID of version to get.
        :return: Version object for given version id.
        """
        return self.evg_api.version_by_id(version_id)

    def query_version_existence(self, version_id: str) -> bool:
        """
        Query if the given evergreen version exists.

        :param version_id: ID of evergreen version to query.
        :returns: True if the version exists.
        """
        try:
            version = self.evg_api.version_by_id(version_id)
            return version is not None
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            LOGGER.warning("Unexpected error returned by Evergreen.", exc_info=True)
            raise err

    def get_mongo_projects(self) -> List[str]:
        """Return the list of mongodb/mongo evergreen project identifiers."""
        evg_projects = [
            proj
            for proj in self.evg_api.all_projects()
            if proj.enabled and MONGO_PROJECT_PREFIX in proj.identifier
        ]
        return [proj.identifier for proj in evg_projects]

    def get_evergreen_version(
        self, evergreen_projects: List[str], evg_ref: str
    ) -> Optional[Version]:
        """Return evergreen version by reference (commit_hash or evergreen_version_id)."""
        # Evergreen reference as evergreen_version_id
        evg_refs = [evg_ref]
        # Evergreen reference as {project_name}_{commit_hash}
        evg_refs.extend(f"{proj.replace('-', '_')}_{evg_ref}" for proj in evergreen_projects)

        for ref in evg_refs:
            try:
                evg_version = self.evg_api.version_by_id(ref)
            except HTTPError:
                continue
            else:
                LOGGER.debug(
                    "Found evergreen version.",
                    evergreen_version=evg_version.version_id,
                )
                return evg_version

        return None

    def get_version_iterator(self, evg_project: str) -> Iterator[Version]:
        """
        Get an iterator over evergreen versions for the given project.

        :param evg_project: Evergreen project to query.
        :return: Iterator over the versions of the given evergreen project.
        """
        return self.evg_api.versions_by_project(evg_project)

    def get_compile_artifact_urls(
        self, evg_version: Version, buildvariant_name: str, ignore_failed_push: bool = False
    ) -> Optional[ArtifactUrls]:
        """
        Find the artifact URLs for the given build variant.

        :param evg_version: Evergreen version to search.
        :param buildvariant_name: Name of build variant to query.
        :param ignore_failed_push: Return URLs even if the push task failed.
        :return: URLs to artifacts in build variant.
        """
        build_id = evg_version.build_variants_map[buildvariant_name]
        evg_build = self.evg_api.build_by_id(build_id)
        LOGGER.debug("Checking build", build_id=evg_build.id)

        evg_tasks = evg_build.get_tasks()
        required_tasks = RequiredTasks.from_task_list(
            evg_tasks,
            self.setup_repro_config.get_evergreen_tasks(evg_version.project_identifier),
            not ignore_failed_push,
        )
        if required_tasks is not None:
            return required_tasks.get_artifact_urls()
        return None



"""Common config module."""
from __future__ import annotations

import os
from enum import Enum
from typing import List, NamedTuple, Optional, Set

from evergreen import Task
from pydantic import BaseModel

from db_contrib_tool.utils.filesystem import read_yaml_file

CONFIG_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "config")
SETUP_REPRO_ENV_CONFIG_FILE = os.path.join(CONFIG_DIR, "setup_repro_env_config.yml")
# Records the paths of installed multiversion binaries on Windows.
WINDOWS_BIN_PATHS_FILE = "windows_binary_paths.txt"
DEFAULT_VERSION_LABEL = "default"


class SegmentWriteKey(Enum):
    DEV = "RmMosrk1bf025xLRbZUxF2osVzGLPpL3"
    PROD = "jrHGRmtQBO8HYU1CyCfnB279SnktLgGH"


SEGMENT_WRITE_KEY = SegmentWriteKey.PROD.value
if os.environ.get("ENV") == "DEV":
    SEGMENT_WRITE_KEY = SegmentWriteKey.DEV.value


class DownloadTarget(NamedTuple):
    """
    Build variant attributes to target for download.

    :param edition: Edition of mongo (e.g. enterprise).
    :param platform: Platform mongo was built for (e.g. ubuntu1804).
    :param architecture: CPU architecture mongo was built for (e.g. x86).
    """

    edition: str
    platform: Optional[str]
    architecture: str

    def is_complete(self) -> bool:
        """Determine if this target has all the details it needs."""
        return self.platform is not None


class Tasks(BaseModel):
    """
    Important evergreen tasks for setup repro env config.

    * projects: Evergreen projects that these tasks apply to.
    * binary: Set of tasks that create mongo binaries.
    * symbols: Set of tasks that publish debug symbols.
    * push: Set of tasks that push artifacts.
    """

    projects: Set[str]
    binary: Set[str]
    symbols: Set[str]
    push: Set[str]

    def is_task_binary(self, evg_task: Task) -> bool:
        """Determine if the given evergreen task generates mongo binaries."""
        return evg_task.display_name in self.binary

    def is_task_symbols(self, evg_task: Task) -> bool:
        """Determine if the given evergreen task generates debug symbols."""
        return evg_task.display_name in self.symbols

    def is_task_push(self, evg_task: Task) -> bool:
        """Determine if the given evergreen task pushes artifacts."""
        return evg_task.display_name in self.push


class Buildvariant(BaseModel):
    """Class represents evergreen buildvariant in setup repro env config."""

    name: str
    edition: str
    platform: str
    architecture: str
    versions: Optional[Set[str]] = None

    def is_match(self, target: DownloadTarget) -> bool:
        """
        Determine if this build variant matches the given criteria.

        :param target: Criteria to match build variant.
        :returns: True if all the criteria match.
        """
        return all(
            [
                target.edition == self.edition,
                target.platform == self.platform,
                target.architecture == self.architecture,
            ]
        )

    def targets_version(self, version: str) -> bool:
        """
        Determine if this build variant targets the given version.

        :param version: Version to check against.
        :returns: True if the given version is targeted.
        """
        return self.versions is not None and version in self.versions


class SetupReproEnvConfig(BaseModel):
    """Class represents setup repro env config."""

    evergreen_tasks: List[Tasks]
    evergreen_buildvariants: List[Buildvariant]

    @classmethod
    def from_yaml_file(cls, filename: str) -> SetupReproEnvConfig:
        """Create an instance of the class from the given filename."""
        return cls(**read_yaml_file(filename))

    @classmethod
    def empty_config(cls) -> SetupReproEnvConfig:
        """Create an empty instance of the configuration."""
        return cls(
            evergreen_tasks=[
                Tasks(
                    projects=set(),
                    binary=set(),
                    symbols=set(),
                    push=set(),
                )
            ],
            evergreen_buildvariants=[],
        )

    def find_matching_build_variant(
        self, target: DownloadTarget, target_version: str
    ) -> Optional[str]:
        """
        Find a build variant that matches the given criteria.

        :param target: Criteria of build variant to target.
        :param target_version: Version to target.
        :return: Name of build variant matching given criteria.
        """
        last_found_name = None
        evergreen_buildvariants = self.evergreen_buildvariants

        for buildvariant in evergreen_buildvariants:
            if buildvariant.is_match(target):
                if buildvariant.targets_version(target_version):
                    return buildvariant.name
                elif not buildvariant.versions:
                    last_found_name = buildvariant.name

        return last_found_name

    def get_evergreen_tasks(self, evg_project: str) -> Tasks:
        """
        Find the set of required tasks that match the given evg_project.

        :param evg_project: Evergreen project to search for.
        :return: Required tasks for the evergreen project.
        """
        default_tasks = None
        for tasks in self.evergreen_tasks:
            if evg_project in tasks.projects:
                return tasks
            if DEFAULT_VERSION_LABEL in tasks.projects:
                default_tasks = tasks

        if default_tasks is None:
            raise ValueError("Default tasks not found in config.")
        return default_tasks


SETUP_REPRO_ENV_CONFIG = SetupReproEnvConfig.from_yaml_file(SETUP_REPRO_ENV_CONFIG_FILE)



"""Client for working with input/output."""
import sys


class IOClient:
    """Client for working with input/output."""

    @staticmethod
    def read_from_stdin() -> str:
        """
        Read from standard input.

        :return: Standard input string.
        """
        return sys.stdin.read()



"""Service for performing git workflows."""
from typing import Optional

import inject
import structlog

from db_contrib_tool.clients.git_client import GitClient

LOGGER = structlog.get_logger(__name__)


class GitService:
    """A service for performing git workflows."""

    @inject.autoparams()
    def __init__(self, git_client: GitClient) -> None:
        """
        Initialize the service.

        :param git_client: Client for interacting with git.
        """
        self.git_client = git_client

    def get_merge_base_commit(self, mongo_version: str) -> Optional[str]:
        """
        Get the merge-base between origin/master and the given version.

        :param version: Mongo version to find mergebase of.
        :return: Commit hash of merge-base.
        """
        target_branch = f"origin/v{mongo_version}"
        result = self.git_client.merge_base("origin/master", target_branch)
        if result.return_code != 0:
            current_branch_result = self.git_client.rev_parse("HEAD", abbrev_ref=True)
            if current_branch_result.output == "master":
                LOGGER.warning(
                    "Git command failed.",
                    cmd="merge-base",
                    branch=target_branch,
                    error=result.error,
                )
                LOGGER.warning("Falling back to the latest starting from the current commit.")

                result = self.git_client.rev_parse("HEAD", verify=True)

        if result.return_code != 0:
            LOGGER.warning(
                "Could not find mergebase",
                error=result.error,
            )
            LOGGER.warning("Falling back to the latest.")
            return None

        commit_hash = result.output
        LOGGER.info("Found commit.", commit=commit_hash)
        return commit_hash

    def get_commit_from_tag(self, tag: str) -> Optional[str]:
        """
        Get the commit hash that git tag points to.

        :param tag: Git tag.
        :return: Commit hash.
        """
        result = self.git_client.rev_list(tag, 1)
        if result.return_code != 0:
            LOGGER.warning(
                "Git command failed.",
                cmd="rev-parse",
                error=result.error,
            )
            return None

        commit_hash = result.output
        LOGGER.info("Found commit.", commit=commit_hash)
        return commit_hash



"""Usage data collecting and storing in external sources."""
import os
import uuid
from typing import Any, Dict, Optional

import analytics
import click
import yaml
from pydantic import BaseModel

from db_contrib_tool.config import SEGMENT_WRITE_KEY


def _get_username() -> str:
    """Get username from .evergreen.yml."""
    try:
        with open(os.path.expanduser(os.path.join("~", ".evergreen.yml")), "r") as fh:
            evg_yml_data = yaml.safe_load(fh)
    except FileNotFoundError:
        evg_yml_data = {}
    return evg_yml_data.get("user", "")


USERNAME = _get_username()
ANONYMOUS_ID = str(uuid.uuid4())
analytics.write_key = SEGMENT_WRITE_KEY


class CommandUsage(BaseModel):
    """
    Command usage information.

    * command: Name of subcommand being invoked.
    """

    command: Optional[str]


class CommandWithUsageTracking(click.Command):
    """Class to track click command usage."""

    def invoke(self, ctx) -> Any:
        """
        Given a context, this invokes the attached callback (if it exists) in the right way.

        :param ctx: Invocation context for this command.
        :return: Invocation result.
        """
        ctx.ensure_object(CommandUsage)
        track_usage(
            event="Command Line Tool Invoked",
            properties={
                "command": ctx.obj.command,
                "parsed_arguments": ctx.params,
            },
        )
        return super().invoke(ctx)


def track_usage(event: str, properties: Dict[str, Any]) -> None:
    """Track event.

    :param event: use Title Case event names, e.g. `Event Name`
    :param properties: use snake_case property keys, e.g. `property_name`,
        and serializable to JSON values
    """

    # Function call does not result in an HTTP request, but is queued in memory instead.
    # Messages are flushed in batch in the background
    # https://segment.com/docs/connections/sources/catalog/libraries/server/python/#batching
    analytics.track(
        user_id="",
        event=event,
        properties=properties,
        context={
            "traits": {
                "username": USERNAME,
            },
        },
        anonymous_id=ANONYMOUS_ID,
    )



"""A client to proxy platform details."""

import platform

import distro


class PlatformDetailsClient:
    """A client to get details about the platform being run on."""

    def get_platform_system(self) -> str:
        """Get the system that is being run on."""
        return platform.system()

    def get_distro_id(self) -> str:
        """Get the id of the distro being run on."""
        return distro.id()

    def get_distro_major_version(self) -> str:
        """Get the major version of the distro being run on."""
        return distro.major_version()

    def get_distro_minor_version(self) -> str:
        """Get the minor version of the distro being run on."""
        return distro.minor_version()



"""Setup MongoDB repro environment."""
import os
import re
from typing import List, NamedTuple, Optional

import inject
import structlog

from db_contrib_tool.clients.download_client import DownloadError
from db_contrib_tool.clients.file_service import FileService
from db_contrib_tool.clients.resmoke_proxy import ResmokeProxy
from db_contrib_tool.config import WINDOWS_BIN_PATHS_FILE, DownloadTarget
from db_contrib_tool.services.evergreen_service import EvergreenService
from db_contrib_tool.setup_repro_env.artifact_discovery_service import (
    ArtifactDiscoveryService,
    RequestTarget,
    RequestType,
)
from db_contrib_tool.setup_repro_env.download_service import (
    ArtifactDownloadService,
    DownloadOptions,
)
from db_contrib_tool.utils import evergreen_conn, is_windows

BINARY_ARTIFACT_NAME = "Binaries"
KNOWN_BRANCHES = {"master"}
RELEASE_VERSION_RE = re.compile(r"^\d+\.\d+$")
PATCH_VERSION_RE = re.compile(r"^\d+\.\d+\.\d+")
BRANCH_RE = re.compile(r"^v\d+\.\d+")
EXTERNAL_LOGGERS = [
    "evergreen",
    "github",
    "inject",
    "segment",
    "urllib3",
]

LOGGER = structlog.getLogger(__name__)


class SetupReproParameters(NamedTuple):
    """
    Parameters describing how a repro environment should be setup.

    * edition: MongoDB edition to download.
    * platform: Target platform to download.
    * architecture: Target architecture to download.
    * variant: Build Variant to download from.

    * versions: List of items to download.
    * install_last_lts: If True download last LTS version of mongo.
    * install_last_continuous: If True download last continuous version of mongo.
    * ignore_failed_push: Download version even if the push task failed.
    * fallback_to_master: Should the latest master be downloaded if the version doesn't exist.

    * evg_version_file: Write which evergreen version were downloaded from to this file.

    * download_options: Options specifying how downloads should occur.
    """

    edition: str
    platform: str
    architecture: str
    variant: str

    versions: List[str]
    install_last_lts: bool
    install_last_continuous: bool
    ignore_failed_push: bool
    fallback_to_master: bool

    evg_version_file: Optional[str]

    download_options: DownloadOptions

    def get_download_target(self, platform: Optional[str] = None) -> DownloadTarget:
        """
        Get the download target to use based on these parameters.

        :param platform: Override the platform with this platform.
        :return: Download target specified by this options.
        """
        platform = platform if platform is not None else self.platform
        return DownloadTarget(
            edition=self.edition, platform=platform, architecture=self.architecture
        )


class SetupReproOrchestrator:
    """Orchestrator for setting up repro environments."""

    @inject.autoparams()
    def __init__(
        self,
        evg_service: EvergreenService,
        resmoke_proxy: ResmokeProxy,
        artifact_download_service: ArtifactDownloadService,
        artifact_discovery_service: ArtifactDiscoveryService,
        file_service: FileService,
    ) -> None:
        """
        Initialize the orchestrator.

        :param evg_service: Service for working with evergreen.
        :param resmoke_proxy: Proxy for working with resmoke.
        :param artifact_download_service: Service to download artifacts.
        :param artifact_discovery_service: Service to find artifacts.
        :param file_service: Service to work with the filesystem.
        """
        self.evg_service = evg_service
        self.resmoke_proxy = resmoke_proxy
        self.artifact_download_service = artifact_download_service
        self.artifact_discovery_service = artifact_discovery_service
        self.file_service = file_service

    def interpret_request(self, request: str) -> RequestTarget:
        """
        Translate the request from the user into an item we can understand.

        :param request: Request from user.
        :return: Targeted request to download.
        """
        if request in KNOWN_BRANCHES or BRANCH_RE.match(request):
            return RequestTarget(RequestType.GIT_BRANCH, request)

        if RELEASE_VERSION_RE.match(request):
            return RequestTarget(RequestType.MONGO_RELEASE_VERSION, request)

        if PATCH_VERSION_RE.match(request):
            return RequestTarget(RequestType.MONGO_PATCH_VERSION, request)

        if self.evg_service.query_task_existence(request):
            return RequestTarget(RequestType.EVG_TASK, request)

        if self.evg_service.query_version_existence(request):
            return RequestTarget(RequestType.EVG_VERSION, request)

        return RequestTarget(RequestType.GIT_COMMIT, request)

    def interpret_requests(
        self, request_list: List[str], last_lts: bool, last_continuous: bool
    ) -> List[RequestTarget]:
        """
        Translate all the requests from the user into items we can understand.

        :param request_list: Requests from user.
        :param last_lts: Should 'last lts' version be included.
        :param last_continuous: Should the 'last continuous' version be included.
        :return: List of targeted request to download.
        """
        requests = [self.interpret_request(request) for request in request_list]
        if last_lts or last_continuous:
            requests.extend(self._get_release_versions(last_lts, last_continuous))

        return requests

    def _get_release_versions(
        self, install_last_lts: Optional[bool], install_last_continuous: Optional[bool]
    ) -> List[RequestTarget]:
        """
        Create a list of multiversion versions that should be included.

        :param install_last_lts: True if the last LTS version should be included.
        :param install_last_continuous: True if the last continuous version should be included.
        :return: List of which multiversion versions should be included.
        """
        multiversionconstants = self.resmoke_proxy.get_multiversion_constants()
        releases = {
            multiversionconstants.last_lts_fcv: install_last_lts,
            multiversionconstants.last_continuous_fcv: install_last_continuous,
        }
        LOGGER.debug("LTS and continuous release inclusions", releases=releases)
        out = {
            RequestTarget.previous_release(version)
            for version, requested in releases.items()
            if requested
        }

        return list(out)

    @staticmethod
    def _get_bin_suffix(version: str, evg_project_id: str) -> str:
        """Get the multiversion bin suffix from the evergreen project ID."""
        if re.match(r"(\d+\.\d+)", version):
            # If the cmdline version is already a semvar, just use that.
            return version
        elif evg_project_id in ("mongodb-mongo-master", "mongodb-mongo-master-nightly"):
            # If the version is not a semvar and the project is the master waterfall,
            # we can't add a suffix.
            return ""
        else:
            # Use the Evergreen project ID as fallback.
            return re.search(r"(\d+\.\d+$)", evg_project_id).group(0)

    def execute(self, setup_repro_params: SetupReproParameters) -> bool:
        """Execute setup repro env mongodb."""
        request_list = self.interpret_requests(
            setup_repro_params.versions,
            setup_repro_params.install_last_lts,
            setup_repro_params.install_last_continuous,
        )

        downloaded_versions = []
        failed_requests = []
        link_directories = []

        download_target = setup_repro_params.get_download_target()
        LOGGER.info("Search criteria", search_criteria=download_target)

        for request in request_list:
            LOGGER.info("Setting up request", request=request)
            LOGGER.info("Fetching download URLs from Evergreen")

            try:
                urls_info = self.artifact_discovery_service.find_artifacts(
                    request,
                    setup_repro_params.variant,
                    download_target,
                    setup_repro_params.ignore_failed_push,
                    setup_repro_params.fallback_to_master,
                )

                if urls_info is None:
                    failed_requests.append(request)
                    LOGGER.warning("Unable to find artifacts for request", request=request)
                    continue

                bin_suffix = self._get_bin_suffix(request.identifier, urls_info.project_identifier)
                linked_dir = self.artifact_download_service.download_and_extract(
                    urls_info.urls,
                    bin_suffix,
                    urls_info.evg_version_id,
                    setup_repro_params.download_options,
                )
                if linked_dir:
                    link_directories.append(linked_dir)
                downloaded_versions.append(urls_info.evg_version_id)
                LOGGER.info("Setup request completed", request=request)
            except (
                evergreen_conn.EvergreenConnError,
                DownloadError,
            ):
                failed_requests.append(request)
                LOGGER.error("Setup request failed", request=request, exc_info=True)

        if is_windows():
            self.file_service.write_windows_install_paths(WINDOWS_BIN_PATHS_FILE, link_directories)

        if setup_repro_params.evg_version_file is not None:
            self.file_service.append_lines_to_file(
                setup_repro_params.evg_version_file, downloaded_versions
            )
            LOGGER.info(
                "Finished writing downloaded Evergreen versions",
                target_file=os.path.abspath(setup_repro_params.evg_version_file),
            )

        if len(downloaded_versions) < len(request_list):
            LOGGER.error("Some requests were not able to setup.", failed_requests=failed_requests)
            return False
        LOGGER.info("Downloaded versions", request_list=request_list)
        return True



"""A client to download artifacts."""
import contextlib
import ctypes
import errno
import os
import shutil
import tarfile
import zipfile
from typing import Optional

import requests
import structlog

from db_contrib_tool.utils.filesystem import build_bin_path, mkdtemp_in_build_dir

LOGGER = structlog.getLogger(__name__)


class DownloadError(Exception):
    """Errors resulting from download failures."""

    pass


class DownloadClient:
    """A client to download artifacts."""

    def download_from_s3(self, url: str) -> str:
        """
        Download a file from the given URL pointing to an S3 bucket.

        :param url: URL of file to download.
        :return: Path to downloaded file on local filesystem.
        """
        if not url:
            raise DownloadError("Download URL not found")

        LOGGER.info("Downloading", url=url)
        filename = os.path.join(mkdtemp_in_build_dir(), url.split("/")[-1].split("?")[0])

        with requests.get(url, stream=True) as reader:
            with open(filename, "wb") as file_handle:
                shutil.copyfileobj(reader.raw, file_handle)

        return filename

    def extract_archive(self, archive_file: str, install_dir: str) -> str:
        """
        Extract the given archive files content into the specified directory.

        :param archive_file: Path to file to extract.
        :param install_dir: Path of directory to extract into.
        :return: Path to directory file were extracted into.
        """
        LOGGER.info("Extracting archive data.", archive=archive_file, install_dir=install_dir)
        temp_dir = mkdtemp_in_build_dir()
        archive_name = os.path.basename(archive_file)
        _, file_suffix = os.path.splitext(archive_name)

        if file_suffix == ".zip":
            # Support .zip downloads, used for Windows binaries.
            with zipfile.ZipFile(archive_file) as zip_handle:
                zip_handle.extractall(temp_dir)
        elif file_suffix == ".tgz":
            # Support .tgz downloads, used for Linux binaries.
            with contextlib.closing(tarfile.open(archive_file, "r:gz")) as tar_handle:
                tar_handle.extractall(path=temp_dir)
        else:
            raise DownloadError(f"Unsupported file extension {file_suffix}")

        try:
            os.makedirs(install_dir)
        except FileExistsError:
            pass

        _rsync_move_dir(temp_dir, install_dir)
        shutil.rmtree(temp_dir)

        LOGGER.info("Extract archive completed.", installed_dir=install_dir)

        return install_dir

    def symlink_version(
        self, suffix: str, installed_dir: str, link_dir: Optional[str] = None
    ) -> str:
        """
        Symlink the binaries in the 'installed_dir' to the 'link_dir'.

        If `link_dir` is None, link to the physical executable's directory (`bin_dir`).
        """
        bin_dir = build_bin_path(installed_dir)
        if os.path.isdir(bin_dir):
            bin_dir = bin_dir
        else:
            bin_dir = installed_dir

        if link_dir is None:
            link_dir = bin_dir
        else:
            mkdir_p(link_dir)

        for executable in os.listdir(bin_dir):
            if executable.endswith(".dll"):
                LOGGER.debug("Skipping linking DLL", file=executable)
                continue

            executable_name, executable_extension = os.path.splitext(executable)
            if suffix:
                link_name = f"{executable_name}-{suffix}{executable_extension}"
            else:
                link_name = executable

            executable = os.path.join(bin_dir, executable)
            executable_link = os.path.join(link_dir, link_name)

            create_symlink(executable, executable_link)

        LOGGER.info("Symlinks for all executables are created in the directory.", link_dir=link_dir)
        return link_dir


def _rsync_move_dir(source_dir, dest_dir):
    """
    Move dir.

    Move the contents of `source_dir` into `dest_dir` as a subdir while merging with
    all existing dirs.

    This is similar to the behavior of `rsync` but different to `mv`.
    """

    for cur_src_dir, _, files in os.walk(source_dir):
        cur_dest_dir = cur_src_dir.replace(source_dir, dest_dir, 1)
        if not os.path.exists(cur_dest_dir):
            os.makedirs(cur_dest_dir)
        for cur_file in files:
            src_file = os.path.join(cur_src_dir, cur_file)
            dst_file = os.path.join(cur_dest_dir, cur_file)
            if os.path.exists(dst_file):
                # in case of the src and dst are the same file
                if os.path.samefile(src_file, dst_file):
                    continue
                os.remove(dst_file)
            shutil.move(src_file, cur_dest_dir)


def mkdir_p(path):
    """Python equivalent of `mkdir -p`."""
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def create_symlink(executable, executable_link):
    """Symlink executable to executable_link."""
    link_method = os.symlink
    if os.name == "nt":
        # os.symlink is not supported on Windows, use a direct method instead.
        def symlink_ms(source, symlink_name):
            """Provide symlink for Windows."""
            csl = ctypes.windll.kernel32.CreateSymbolicLinkW
            csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
            csl.restype = ctypes.c_ubyte
            flags = 1 if os.path.isdir(source) else 0
            if csl(symlink_name, source.replace("/", "\\"), flags) == 0:
                raise ctypes.WinError()

        link_method = symlink_ms

    try:
        link_method(executable, executable_link)
        LOGGER.debug("Symlink created.", executable=executable, executable_link=executable_link)

    except OSError as exc:
        if exc.errno == errno.EEXIST:
            LOGGER.warning("Symlink already exists.", exc=exc)
            if os.name != "nt":
                LOGGER.warning(
                    "Removing old symlink & trying again", executable_link=executable_link
                )
                os.remove(executable_link)
                link_method(executable, executable_link)
                LOGGER.debug(
                    "Symlink created.", executable=executable, executable_link=executable_link
                )
            pass
        else:
            raise



"""A service for working with platforms."""
from typing import Optional

import inject

from db_contrib_tool.clients.platform_details import PlatformDetailsClient
from db_contrib_tool.config import SETUP_REPRO_ENV_CONFIG_FILE


class PlatformService:
    """A service for working with platforms."""

    @inject.autoparams()
    def __init__(self, platform_details_client: PlatformDetailsClient) -> None:
        """
        Initialize the service.

        :param platform_details_client: Client for getting platform details.
        """
        self.platform_details_client = platform_details_client

    def infer_platform(self, edition: Optional[str] = None, version: Optional[str] = None) -> str:
        """
        Determine which platform should be downloaded based on the system being run on.

        :param edition: Mongo edition that should be targeted.
        :param version: Mongo version that should be targeted.
        :return: Platform to target.
        """
        system_name = self.platform_details_client.get_platform_system()
        platform_name = None

        if system_name == "Darwin":
            platform_name = "osx"

        elif system_name == "Windows":
            platform_name = "windows"
            if edition == "base" and version == "4.2":
                platform_name += "_x86_64-2012plus"

        elif system_name == "Linux":
            id_name = self.platform_details_client.get_distro_id()
            if id_name in ("ubuntu", "rhel"):
                platform_name = (
                    id_name
                    + self.platform_details_client.get_distro_major_version()
                    + self.platform_details_client.get_distro_minor_version()
                )

        if platform_name is None:
            raise ValueError(
                "Platform cannot be inferred. Please specify platform explicitly with -p. "
                f"Available platforms can be found in {SETUP_REPRO_ENV_CONFIG_FILE}."
            )
        else:
            return platform_name



"""Command-line entry-point into db-contrib-tool."""

import click

from db_contrib_tool.evg_aware_bisect.cli import bisect
from db_contrib_tool.setup_repro_env.cli import setup_repro_env
from db_contrib_tool.symbolizer.cli import symbolize
from db_contrib_tool.usage_analytics import CommandUsage

_PLUGINS = [
    setup_repro_env,
    symbolize,
    bisect,
]


@click.group(context_settings=dict(show_default=True, max_content_width=120))
@click.pass_context
def cli(ctx: click.Context) -> None:
    """
    The db-contrib-tool - MongoDB's tool for contributors.

    For more information, see the help message for each subcommand.
    For example: db-contrib-tool setup-repro-env --help
    """
    ctx.obj = CommandUsage(command=ctx.invoked_subcommand)


for plugin in _PLUGINS:
    cli.add_command(plugin)


if __name__ == "__main__":
    cli(obj=CommandUsage())



"""Helper tools to get OAuth credentials using the PKCE flow."""
from __future__ import annotations

from datetime import datetime, timedelta
from http.server import BaseHTTPRequestHandler, HTTPServer
from random import choice
from string import ascii_lowercase
from typing import Any, Callable, Optional, Tuple
from urllib.parse import parse_qs, urlsplit
from webbrowser import open as web_open

import requests
from oauthlib.oauth2 import BackendApplicationClient
from pkce import generate_pkce_pair
from pydantic import ValidationError
from pydantic.main import BaseModel
from requests_oauthlib import OAuth2Session

from db_contrib_tool.utils.filesystem import read_yaml_file

AUTH_HANDLER_RESPONSE = """\
<html>
  <head>
    <title>Authentication Status</title>
    <script>
    window.onload = function() {
      window.close();
    }
    </script>
  </head>
  <body>
    <p>The authentication flow has completed.</p>
  </body>
</html>
""".encode(
    "utf-8"
)


class Configs:
    """Collect configurations necessary for authentication process."""

    AUTH_DOMAIN = "corp.mongodb.com/oauth2/aus4k4jv00hWjNnps297"
    CLIENT_ID = "0oa5zf9ps4N3JKWIJ297"
    REDIRECT_PORT = 8989
    SCOPE = "kanopy+openid+profile"

    def __init__(
        self,
        client_credentials_scope: str = None,
        client_credentials_user_name: str = None,
        auth_domain: str = None,
        client_id: str = None,
        redirect_port: int = None,
        scope: str = None,
    ):
        """Initialize configs instance."""

        self.AUTH_DOMAIN = auth_domain or self.AUTH_DOMAIN
        self.CLIENT_ID = client_id or self.CLIENT_ID
        self.REDIRECT_PORT = redirect_port or self.REDIRECT_PORT
        self.SCOPE = scope or self.SCOPE
        self.CLIENT_CREDENTIALS_SCOPE = client_credentials_scope
        self.CLIENT_CREDENTIALS_USER_NAME = client_credentials_user_name


class OAuthCredentials(BaseModel):
    """OAuth access token and its associated metadata."""

    expires_in: int
    access_token: str
    created_time: datetime
    user_name: str

    def are_expired(self) -> bool:
        """
        Check whether the current OAuth credentials are expired or not.

        :return: Whether the credentials are expired or not.
        """
        return self.created_time + timedelta(seconds=self.expires_in) < datetime.now()

    @classmethod
    def get_existing_credentials_from_file(cls, file_path: str) -> Optional[OAuthCredentials]:
        """
        Try to get OAuth credentials from a file location.

        Will return None if credentials either don't exist or are expired.
        :param file_path: Location to check for OAuth credentials.
        :return: Valid OAuth credentials or None if valid credentials don't exist
        """
        try:
            creds = OAuthCredentials(**read_yaml_file(file_path))
            if (
                creds.access_token
                and creds.created_time
                and creds.expires_in
                and creds.user_name
                and not creds.are_expired()
            ):
                return creds
            else:
                return None
        except ValidationError:
            return None
        except OSError:
            return None


class _RedirectServer(HTTPServer):
    """HTTP server to use when fetching OAuth credentials using the PKCE flow."""

    pkce_credentials: Optional[OAuthCredentials] = None
    auth_domain: str
    client_id: str
    redirect_uri: str
    code_verifier: str

    def __init__(
        self,
        server_address: Tuple[str, int],
        handler: Callable[..., BaseHTTPRequestHandler],
        redirect_uri: str,
        auth_domain: str,
        client_id: str,
        code_verifier: str,
    ):
        self.redirect_uri = redirect_uri
        self.auth_domain = auth_domain
        self.client_id = client_id
        self.code_verifier = code_verifier
        super().__init__(server_address, handler)


class _Handler(BaseHTTPRequestHandler):
    """Request handler class to use when trying to get OAuth credentials."""

    server: _RedirectServer

    def _set_response(self) -> None:
        """Set the response to the server making a request."""
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def log_message(self, log_format: Any, *args: Any) -> None:
        """
        Log HTTP Server internal messages.

        :param log_format: The format to use when logging messages.
        :param args: Key word args.
        """
        return None

    def do_GET(self) -> None:
        """Handle the callback response from the auth server."""
        params = parse_qs(urlsplit(self.path).query)
        code = params.get("code")

        if not code:
            raise ValueError("Could not get authorization code when signing in to Okta")

        url = f"https://{self.server.auth_domain}/v1/token"
        body = {
            "grant_type": "authorization_code",
            "client_id": self.server.client_id,
            "code_verifier": self.server.code_verifier,
            "code": code,
            "redirect_uri": self.server.redirect_uri,
        }

        resp = requests.post(url, data=body).json()

        access_token = resp.get("access_token")
        expires_in = resp.get("expires_in")

        if not access_token or not expires_in:
            raise ValueError("Could not get access token or expires_in data about access token")

        headers = {"Authorization": f"Bearer {access_token}"}
        resp = requests.get(
            f"https://{self.server.auth_domain}/v1/userinfo", headers=headers
        ).json()

        split_username = resp["preferred_username"].split("@")

        if len(split_username) != 2:
            raise ValueError("Could not get user_name of current user")

        self.server.pkce_credentials = OAuthCredentials(
            access_token=access_token,
            expires_in=expires_in,
            created_time=datetime.now(),
            user_name=split_username[0],
        )
        self._set_response()
        self.wfile.write(AUTH_HANDLER_RESPONSE)


class PKCEOauthTools:
    """Basic toolset to get OAuth credentials using the PKCE flow."""

    auth_domain: str
    client_id: str
    redirect_port: int
    redirect_uri: str
    scope: str

    def __init__(self, auth_domain: str, client_id: str, redirect_port: int, scope: str):
        """
        Create a new PKCEOauth tools instance.

        :param auth_domain: The uri of the auth server to get the credentials from.
        :param client_id: The id of the client that you are using to authenticate.
        :param redirect_port: Port to use when setting up the local server for the auth redirect.
        :param scope: The OAuth scopes to request access for.
        """
        self.auth_domain = auth_domain
        self.client_id = client_id
        self.redirect_port = redirect_port
        self.redirect_uri = f"http://localhost:{redirect_port}/"
        self.scope = scope

    def get_pkce_credentials(self, print_auth_url: bool = False) -> OAuthCredentials:
        """
        Try to get an OAuth access token and its associated metadata.

        :param print_auth_url: Whether to print the auth url to the console instead of opening it.
        :return: OAuth credentials and some associated metadata to check if they have expired.
        """
        code_verifier, code_challenge = generate_pkce_pair()

        state = "".join(choice(ascii_lowercase) for i in range(10))

        authorization_url = (
            f"https://{self.auth_domain}/v1/authorize?"
            f"scope={self.scope}&"
            f"response_type=code&"
            f"response_mode=query&"
            f"client_id={self.client_id}&"
            f"code_challenge={code_challenge}&"
            f"state={state}&"
            f"code_challenge_method=S256&"
            f"redirect_uri={self.redirect_uri}"
        )

        httpd = _RedirectServer(
            ("", self.redirect_port),
            _Handler,
            self.redirect_uri,
            self.auth_domain,
            self.client_id,
            code_verifier,
        )
        if print_auth_url:
            print("Please open the below url in a browser and sign in if necessary")
            print(authorization_url)
        else:
            web_open(authorization_url)
        httpd.handle_request()

        if not httpd.pkce_credentials:
            raise ValueError(
                "Could not retrieve Okta credentials to talk to Kanopy with. "
                "Please sign out of Okta in your browser and try runnning this script again"
            )

        return httpd.pkce_credentials


def get_oauth_credentials(configs: Configs, print_auth_url: bool = False) -> OAuthCredentials:
    """
    Run the OAuth workflow to get credentials for a human user.

    :param configs: Configs instance.
    :param print_auth_url: Whether to print the auth url to the console instead of opening it.
    :return: OAuth credentials for the given user.
    """
    oauth_tools = PKCEOauthTools(
        auth_domain=configs.AUTH_DOMAIN,
        client_id=configs.CLIENT_ID,
        redirect_port=configs.REDIRECT_PORT,
        scope=configs.SCOPE,
    )
    credentials = oauth_tools.get_pkce_credentials(print_auth_url)
    return credentials


def get_client_cred_oauth_credentials(
    client_id: str, client_secret: str, configs: Configs
) -> OAuthCredentials:
    """
    Run the OAuth workflow to get credentials for a machine user.

    :param client_id: The client_id of the machine user to authenticate as.
    :param client_secret: The client_secret of the machine user to authenticate as.
    :param configs: Configs instance.
    :return: OAuth credentials for the given machine user.
    """
    client = BackendApplicationClient(client_id=client_id)
    oauth = OAuth2Session(client=client)
    token = oauth.fetch_token(
        token_url=f"https://{configs.AUTH_DOMAIN}/v1/token",
        client_id=client_id,
        client_secret=client_secret,
        scope=configs.CLIENT_CREDENTIALS_SCOPE,
    )
    access_token = token.get("access_token")
    expires_in = token.get("expires_in")

    if not access_token or not expires_in:
        raise ValueError("Could not get access token or expires_in data about access token")

    return OAuthCredentials(
        access_token=access_token,
        expires_in=expires_in,
        created_time=datetime.now(),
        user_name=configs.CLIENT_CREDENTIALS_USER_NAME,
    )



"""CLI for bisect."""
import os
import sys
from typing import Optional

import click
from click.core import ParameterSource

from db_contrib_tool.evg_aware_bisect import Bisect
from db_contrib_tool.usage_analytics import CommandWithUsageTracking
from db_contrib_tool.utils.evergreen_conn import EVERGREEN_CONFIG_LOCATIONS

DEFAULT_PYTHON_INSTALLATION = "python3"
TOOLCHAIN_PYTHON3 = "/opt/mongodbtoolchain/v3/bin/python3"


@click.command(cls=CommandWithUsageTracking)
@click.pass_context
@click.option(
    "-l",
    "--lookback",
    default=365,
    help="Maximum number of days to look back for versions to test.",
)
@click.option(
    "-b", "--branch", required=True, help="The branch for which versions are being tested."
)
@click.option(
    "-v", "--variant", required=True, help="The variant for which versions are being tested."
)
@click.option(
    "-s",
    "--script",
    type=click.Path(),
    required=True,
    help="Location of the shell test script to run on the versions.",
)
@click.option(
    "-p",
    "--python-installation",
    type=click.Path(),
    default=DEFAULT_PYTHON_INSTALLATION,
    help="Location of a python installation to use for shell commands.",
    show_default=f"`{DEFAULT_PYTHON_INSTALLATION}` or `{TOOLCHAIN_PYTHON3}` (if toolchain is setup) will be used.",
)
@click.option(
    "-ec",
    "--evergreenConfig",
    "evergreen_config",
    type=click.Path(),
    help=f"Location of evergreen configuration file.",
    show_default=f"If not specified it will look for it in the following locations: {EVERGREEN_CONFIG_LOCATIONS}.",
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    help="Set DEBUG logging level.",
)
def bisect(
    ctx: click.Context,
    lookback: int,
    branch: str,
    variant: str,
    script: str,
    python_installation: Optional[str],
    evergreen_config: Optional[str],
    debug: bool,
):
    """
    Perform an evergreen-aware git-bisect to find the 'last passing version' and
    'first failing version' of mongo, with respect to a user provided shell script.

    The 'bisect' command lets a user specify a '--branch', '--variant' & '--lookback' period on which to
    perform a bisect. The user also provides a shell test '--script' which exits with status code 0 to
    indicate a successful test. The command performs the following steps:

    (1) Get all versions for the given '--branch', '--variant' & '--lookback' period from Evergreen.

    (2) Filter the versions for versions that Evergreen has binaries and artifacts for.

    (3) Find the 'middle' version.

    (4) Setup a test environment.

        - The 'build/resmoke-bisect' directory will have a sub directory --
    'build/resmoke-bisect/{version_id}' containing the git repo for this version.

        - The 'binaries' & 'artifacts' will also be downloaded to the directory named
    'build/resmoke-bisect/{version_id}'.

        - Create a virtual environment at 'build/resmoke-bisect/bisect_venv' and
    install packages for this version.

    (5) Activate 'bisect_venv' & run the user provided shell script from within the
    'build/resmoke-bisect/{version_id}' directory.

    (6) Teardown the test environment.

    (7) Repeat steps (3)-(6) on the left half, if (5) failed, or right half, if (5) succeeded.

    This command will print the "Last Known Passing Version" & "First Known Failing Version".

    NOTE: This 'bisect' command assumes a perfect partition between passing & failing versions.
    ie: [Pass, Pass, Pass, Fail, Fail, Fail]
    If there is not a perfect partition, try modifying the '--lookback' period or shell '--script'.
    """

    if ctx.get_parameter_source(
        "python_installation"
    ) == ParameterSource.DEFAULT and os.path.isfile(TOOLCHAIN_PYTHON3):
        python_installation = TOOLCHAIN_PYTHON3

    bisect_cmd = Bisect(
        branch=branch,
        lookback=lookback,
        evergreen_config=evergreen_config,
        variant=variant,
        script=script,
        debug=debug,
        python_installation=python_installation,
    )

    success = bisect_cmd.execute()
    if not success:
        sys.exit(1)



"""A service for downloading artifacts."""

import os
import time
from typing import Dict, List, NamedTuple, Optional

import inject
import structlog

from db_contrib_tool.clients.download_client import DownloadClient, DownloadError
from db_contrib_tool.clients.file_service import FileService
from db_contrib_tool.utils import is_windows

LOGGER = structlog.get_logger(__name__)


class DownloadOptions(NamedTuple):
    """
    Options describing how downloads should occur.

    * download_binaries: Should the binaries be downloaded.
    * download_symbols: Should the debug symbols be downloaded.
    * download_artifacts: Should the build artifacts be downloaded.
    * download_python_venv: Should the python virtualenv be downloaded.
    * install_dir: Directory to install downloaded artifacts.
    * link_dir: Directory to link downloaded files to.
    """

    download_binaries: bool
    download_symbols: bool
    download_artifacts: bool
    download_python_venv: bool

    install_dir: str
    link_dir: str


class ArtifactDownloadService:
    """A service for downloading artifacts."""

    @inject.autoparams()
    def __init__(self, download_client: DownloadClient, file_service: FileService) -> None:
        """
        Initialize the service.

        :param download_client: Client to perform download actions.
        :param file_service: Client to perform filesystem actions.
        """
        self.download_client = download_client
        self.file_service = file_service
        self.is_windows = is_windows()
        self.retry_time_secs = 1.0

    def download_and_extract(
        self,
        urls: Dict[str, str],
        bin_suffix: str,
        path_prefix: str,
        download_options: DownloadOptions,
    ) -> Optional[str]:
        """
        Download and extract artifacts from the given URLs.

        :param urls: URLs with artifacts to download.
        :param bin_suffix: _description_
        :param path_prefix: _description_
        :param download_options: Details about how to download artifacts.
        :return: Directory extracted artifacts where linked to.
        """
        url_list = self.find_urls_to_download(urls, download_options)
        install_dir = os.path.join(download_options.install_dir, path_prefix)
        linked_dir = self.setup_mongodb(
            url_list,
            download_options.download_binaries,
            install_dir,
            download_options.link_dir,
            bin_suffix,
        )
        return linked_dir

    def find_urls_to_download(
        self, urls: Dict[str, str], download_options: DownloadOptions
    ) -> List[str]:
        """
        Collect the urls to download based on the given options.

        :param urls: Known URLs.
        :param download_options: Options describing which URLs to target.
        :return: List of URLs that should be downloaded.
        """
        download_list = []
        if download_options.download_binaries:
            binaries_url = urls.get("Binaries")
            if binaries_url is not None:
                download_list.append(binaries_url)
            else:
                raise DownloadError("Binaries download requested but not URL available")

        if download_options.download_artifacts:
            artifacts_url = urls.get("Artifacts")
            if artifacts_url is not None:
                download_list.append(artifacts_url)
            else:
                raise DownloadError("Evergreen artifacts download requested but not URL available")

        if download_options.download_symbols:
            symbols_url = (
                urls.get(" mongo-debugsymbols.tgz")
                or urls.get("mongo-debugsymbols.tgz")
                or urls.get(" mongo-debugsymbols.zip")
                or urls.get("mongo-debugsymbols.zip")
            )
            if symbols_url is not None:
                download_list.append(symbols_url)
            else:
                raise DownloadError("Symbols download requested but not URL available")

        if download_options.download_python_venv:
            python_venv_url = urls.get("Python venv (see included README.txt)") or urls.get(
                "Python venv (see included venv_readme.txt)"
            )
            if python_venv_url is not None:
                download_list.append(python_venv_url)
            else:
                raise DownloadError("Python venv download requested but not URL available")

        return download_list

    def setup_mongodb(
        self,
        urls: List[str],
        create_symlinks: bool,
        install_dir: str,
        link_dir: str,
        bin_suffix: str,
    ) -> Optional[str]:
        """
        Download artifacts from the given URLs, extract them, and create symlinks.

        :param urls: List of artifact URLs to download.
        :param create_symlinks: Should symlinks be created to downloaded artifacts.
        :param install_dir: Directory to extract artifacts to.
        :param link_dir: Directory to create symlinks in.
        :param bin_suffix: _description_
        :return: Directory symlinks were created in.
        """
        for url in urls:
            try:
                self.try_download(url, install_dir)
            except Exception:
                LOGGER.warning(
                    "Setting up tarball failed with error, retrying once...", exc_info=True
                )
                time.sleep(self.retry_time_secs)
                self.try_download(url, install_dir)

        if create_symlinks:
            if self.is_windows:
                LOGGER.info(
                    "Linking to install_dir on Windows; executable have to live in different "
                    "working directories to avoid DLLs for different versions clobbering each other"
                )
                link_dir = self.download_client.symlink_version(bin_suffix, install_dir, None)
            else:
                link_dir = self.download_client.symlink_version(bin_suffix, install_dir, link_dir)
            return link_dir
        return None

    def try_download(self, target_url: str, install_dir: str) -> None:
        """
        Attempt to download the given URL.

        :param target_url: URL to download.
        :param install_dir: Location to extract the contents of the downloaded URL.
        """
        tarball = self.download_client.download_from_s3(target_url)
        self.download_client.extract_archive(tarball, install_dir)
        self.file_service.delete_file(tarball)



"""Proxy to get data from resmoke."""
from __future__ import annotations

import subprocess
from typing import List, Optional

import structlog
import yaml
from pydantic import BaseModel

LOGGER = structlog.get_logger(__name__)


class MultiversionConfig(BaseModel):
    """
    Multiversion Configuration obtained from resmoke.

    * last_lts_fcv: Version of the last LTS release.
    * last_continuous_fcv: Version of the last continuous release.
    """

    last_lts_fcv: str
    last_continuous_fcv: str


class ResmokeProxy:
    """A proxy for interacting with resmoke."""

    def __init__(self, resmoke_cmd: List[str]) -> None:
        """
        Initialize the service.

        :param resmoke_cmd: Command to invoke resmoke.py.
        """
        self.resmoke_cmd = resmoke_cmd
        self.multiversion_constants: Optional[MultiversionConfig] = None

    @classmethod
    def with_cmd(cls, resmoke_cmd: str) -> ResmokeProxy:
        """
        Create an instance of ResmokeProxy that invokes the given command.

        :param resmoke_cmd: Command to invoke resmoke.
        :return: Instance of ResmokeProxy using the given command.
        """
        return cls([part.strip() for part in resmoke_cmd.split(" ")])

    def _lazy_load(self) -> None:
        """Import multiversionconstants from resmoke."""
        if self.multiversion_constants is None:
            cmd = self.resmoke_cmd
            cmd.append("multiversion-config")
            try:
                result = subprocess.run(" ".join(cmd), capture_output=True, check=True, shell=True)
            except subprocess.CalledProcessError:
                LOGGER.error("Error invoking resmoke", exc_info=True)
                LOGGER.error(
                    "This command should be run from the root of the mongo repo and the resmoke "
                    "virtualenv should be activated."
                )
                LOGGER.error(
                    "If you're running it from the root of the mongo repo and still seeing"
                    " this error, please reach out in #server-testing slack channel."
                )
                raise
            self.multiversion_constants = MultiversionConfig(**yaml.safe_load(result.stdout))

    def get_multiversion_constants(self) -> MultiversionConfig:
        """Get the multiversion constants from resmoke."""
        self._lazy_load()
        assert self.multiversion_constants
        return self.multiversion_constants



"""Interactions with the db-contrib-tool bisect command."""
import logging
import os
import subprocess
import sys
from datetime import datetime, timedelta, timezone

import structlog

from db_contrib_tool.config import SetupReproEnvConfig
from db_contrib_tool.services.evergreen_service import EvergreenService
from db_contrib_tool.utils import evergreen_conn

LOGGER = structlog.getLogger(__name__)

BISECT_DIR = os.path.dirname(os.path.abspath(__file__))
SETUP_TEST_ENV_SH = os.path.join(BISECT_DIR, "setup_test_env.sh")
TEARDOWN_TEST_ENV_SH = os.path.join(BISECT_DIR, "teardown_test_env.sh")
RUN_USER_SCRIPT_SH = os.path.join(BISECT_DIR, "run_user_script.sh")


def setup_logging(debug=False):
    """Enable logging."""
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(
        format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
        level=log_level,
        stream=sys.stdout,
    )
    structlog.configure(logger_factory=structlog.stdlib.LoggerFactory())


class Bisect:
    """Main class for the Bisect subcommand."""

    def __init__(
        self,
        branch,
        lookback,
        variant,
        script,
        python_installation,
        evergreen_config=None,
        debug=None,
    ):
        """Initialize."""
        setup_logging(debug)
        self.branch = branch
        self.lookback = lookback
        self.variant = variant
        self.script = script
        self.python_installation = python_installation
        self.evergreen_config = evergreen_config
        self.evg_api = evergreen_conn.get_evergreen_api(evergreen_config)
        self.evg_service = EvergreenService(self.evg_api, SetupReproEnvConfig.empty_config())

    @staticmethod
    def _teardown_test_env(version):
        """Remove any directories added during testing process."""
        try:
            subprocess.run(["bash", TEARDOWN_TEST_ENV_SH], check=True)
        except subprocess.CalledProcessError as err:
            LOGGER.error("Could not teardown test environment for bisect.", version=version)
            raise err
        LOGGER.info("Completed teardown of test environment for bisect.", version=version)

    def _setup_test_env(self, version):
        """Set up a test environment for the given version."""
        try:
            subprocess.run(
                [
                    "bash",
                    SETUP_TEST_ENV_SH,
                    self.python_installation,
                    self.evergreen_config if self.evergreen_config else "",
                    self.variant,
                    version,
                ],
                check=True,
            )
        except subprocess.CalledProcessError:
            LOGGER.error(
                "Could not setup test environment for bisect -- retrying.", version=version
            )
            try:
                subprocess.run(
                    [
                        "bash",
                        SETUP_TEST_ENV_SH,
                        self.python_installation,
                        self.evergreen_config if self.evergreen_config else "",
                        self.variant,
                        version,
                    ],
                    check=True,
                )
            except subprocess.CalledProcessError as err:
                LOGGER.error("Could not setup test environment for bisect.", version=version)
                raise err
        LOGGER.info("Completed setup of test environment for bisect.", version=version)

    def _run_user_script(self, version):
        """Run the user script in a virtual environment."""
        return subprocess.run(
            ["bash", RUN_USER_SCRIPT_SH, version, self.script], check=False
        ).returncode

    def _test_version_with_script(self, version):
        """Test the given version with the user provided script."""
        self._setup_test_env(version)
        success = self._run_user_script(version)
        self._teardown_test_env(version)
        return success == 0

    def bisect(self, versions):
        """Bisect to find latest passing version assuming ordered by oldest to latest version."""
        if len(versions) == 0:
            return None
        midpoint = len(versions) // 2
        success = self._test_version_with_script(versions[midpoint])
        # if success, keep checking right
        if success:
            LOGGER.info(
                "Version passed user script.",
                version=versions[midpoint],
            )
            return self.bisect(versions[midpoint + 1 :]) or versions[midpoint]
        # if fail, keep checking left
        else:
            LOGGER.info("Version failed user script.", version=versions[midpoint])
            return self.bisect(versions[0:midpoint])

    def find_versions_with_binaries(self):
        """
        Find versions that have binaries for the user provided variant in the lookback period.

        :return: List of versions that have binaries ordered from earliest to latest.
        """
        versions_with_binaries = []
        for version in self.evg_api.versions_by_project_time_window(
            project_id=f'mongodb-mongo-{"" if self.branch == "master" else "v"}{self.branch}',
            before=datetime.now(timezone.utc),
            after=datetime.now(timezone.utc) - timedelta(self.lookback),
        ):
            urls = self.evg_service.get_compile_artifact_urls(
                version, self.variant, ignore_failed_push=True
            )
            if urls is not None and urls.urls.get("Artifacts") and urls.urls.get("Binaries"):
                versions_with_binaries.append(version.revision)

        versions_with_binaries.reverse()
        return versions_with_binaries

    def execute(self) -> bool:
        """
        Perform bisect for the provided branch, variant & lookback period based on result of script.

        Print the last passing version and first failing version.
        """
        versions_with_binaries = self.find_versions_with_binaries()

        # No versions found
        if not versions_with_binaries:
            LOGGER.info(
                "No versions with binaries found for given branch, variant & lookback period.",
                branch=self.branch,
                variant=self.variant,
                lookback=self.lookback,
            )
            return True

        LOGGER.info(
            "Performing bisect on the following versions: ", versions=versions_with_binaries
        )

        last_passing_version = self.bisect(versions_with_binaries)
        first_failing_version = None

        # All versions failed
        if last_passing_version is None:
            first_failing_version = versions_with_binaries[0]
            LOGGER.info("All versions in lookback period failed.")

        # All versions passed
        elif last_passing_version == versions_with_binaries[-1]:
            LOGGER.info("All versions in lookback period passed.")

        else:
            first_failing_version = versions_with_binaries[
                versions_with_binaries.index(last_passing_version) + 1
            ]

        print(f"Last Known Passing Version: {last_passing_version}")
        print(f"First Known Failing Version: {first_failing_version}")
        return True



"""A service for finding artifacts."""
from __future__ import annotations

import re
from enum import Enum
from typing import Dict, List, NamedTuple, Optional

import inject
import structlog
from evergreen import Version
from requests.exceptions import HTTPError

from db_contrib_tool.config import (
    SETUP_REPRO_ENV_CONFIG_FILE,
    DownloadTarget,
    SetupReproEnvConfig,
)
from db_contrib_tool.services.evergreen_service import EvergreenService
from db_contrib_tool.services.git_service import GitService
from db_contrib_tool.services.platform_service import PlatformService

LOGGER = structlog.get_logger(__name__)
BRANCH_FALLBACK = "master"
NON_VERSION_PROJECTS = {"mongodb-mongo-master", "mongodb-mongo-master-nightly"}
PROJECT_PREFIX = "mongodb-mongo-"
VERSION_PROJECT_RE = re.compile(r"mongodb-mongo-v(\d+\.\d+)")


class MissingBuildVariantError(Exception):
    """An error representing missing build variants."""

    pass


class EvgURLInfo(NamedTuple):
    """Wrapper around compile URLs with metadata."""

    urls: Dict[str, str]
    evg_version_id: str
    project_identifier: str


class RequestType(Enum):
    """
    How the mongo binaries are being requested.

    * git_commit: Get binaries from a specific commit.
    * git_branch: Get binaries from a specific git branch.
    * evg_version: Get binaries from a specific evergreen version.
    * evg_task: Get binaries from a specific evergreen task.
    * mongo_release_version: Get binaries for a release version of mongo (e.g. 4.4, 5.0, etc).
    * mongo_patch_version: Get binaries for a patch version of mongo (e.g. 4.2.18, 6.0.0-rc4, 6.1.0-alpha, etc).
    """

    GIT_COMMIT = "git_commit"
    GIT_BRANCH = "git_branch"
    EVG_VERSION = "evg_version"
    EVG_TASK = "evg_task"
    MONGO_RELEASE_VERSION = "mongo_release_version"
    MONGO_PATCH_VERSION = "mongo_patch_version"

    def __str__(self) -> str:
        """Display item as a string."""
        return self.value


class RequestTarget(NamedTuple):
    """
    Version of mongo binaries that was requested.

    * request_type: Type of the identifier.
    * identifier: Identifier used to find request.
    """

    request_type: RequestType
    identifier: str

    @classmethod
    def previous_release(cls, identifier: str) -> RequestTarget:
        """
        Build a target for either the previous LTS or Continuous release.

        :param identifier: Identifier of release to download.
        :return: Request target to download given release.
        """
        if identifier == BRANCH_FALLBACK:
            return cls(request_type=RequestType.GIT_BRANCH, identifier=identifier)
        return cls(request_type=RequestType.MONGO_RELEASE_VERSION, identifier=identifier)

    def __str__(self) -> str:
        """Display item as a string."""
        return f"{self.request_type}({self.identifier})"


class ArtifactDiscoveryService:
    """A service for finding artifacts."""

    @inject.autoparams()
    def __init__(
        self,
        evg_service: EvergreenService,
        platform_service: PlatformService,
        setup_repro_config: SetupReproEnvConfig,
        git_service: GitService,
    ) -> None:
        """
        Initialize the service.

        :param evg_service: Service for interacting with evergreen.
        :param platform_service: Service for discovering platform information.
        :param setup_repro_config: Configuration for setuping up a repro environment.
        :param git_service: Service for performing git workflows.
        """
        self.evg_service = evg_service
        self.platform_service = platform_service
        self.setup_repro_config = setup_repro_config
        self.git_service = git_service

    def find_artifacts(
        self,
        request: RequestTarget,
        requested_variant: Optional[str],
        target: DownloadTarget,
        ignore_failed_push: bool,
        fallback_to_master: bool,
        starting_commit: Optional[str] = None,
    ) -> Optional[EvgURLInfo]:
        """
        Find the artifacts generated for the given request.

        :param request: Request of mongo instance to find.
        :param requested_variant: Name of build variant that was requested.
        :param target: Attributes of the build variant to download from.
        :param ignore_failed_push: Whether a failed push should raise an exception.
        :param fallback_to_master: Fallback to the master branch if binaries were not found.
        :param starting_commit: When scanning multiple commits, start scanning after this commit.
        :return: Links to found artifacts.
        """
        artifact_urls = None
        if request.request_type == RequestType.EVG_VERSION:
            artifact_urls = self.find_artifact_urls_for_version(
                request.identifier, requested_variant, target, ignore_failed_push
            )

        elif request.request_type == RequestType.EVG_TASK:
            task = self.evg_service.get_task(request.identifier)
            if requested_variant is None:
                requested_variant = task.build_variant
            artifact_urls = self.find_artifact_urls_for_version(
                task.version_id, requested_variant, target, ignore_failed_push
            )

        elif request.request_type == RequestType.GIT_BRANCH:
            mongo_project = f"{PROJECT_PREFIX}{request.identifier}"
            try:
                artifact_urls = self.scan_for_artifact_urls(
                    mongo_project, requested_variant, target, ignore_failed_push, starting_commit
                )
            except MissingBuildVariantError:
                if mongo_project in NON_VERSION_PROJECTS:
                    other_project = find_fallback_project(mongo_project)
                    LOGGER.info(
                        "Could not find expected build variant trying other project",
                        mongo_project=mongo_project,
                        other_project=other_project,
                    )
                    artifact_urls = self.scan_for_artifact_urls(
                        other_project,
                        requested_variant,
                        target,
                        ignore_failed_push,
                        starting_commit,
                    )
                else:
                    raise

        elif request.request_type == RequestType.MONGO_RELEASE_VERSION:
            mongo_project = f"{PROJECT_PREFIX}v{request.identifier}"
            try:
                artifact_urls = self.scan_for_artifact_urls(
                    mongo_project, requested_variant, target, ignore_failed_push
                )
            except HTTPError as err:
                if err.response.status_code == 404 and fallback_to_master:
                    LOGGER.warning(
                        "Unable to find requested mongo version, falling back to default branch",
                        requested_version=request.identifier,
                        default_branch=BRANCH_FALLBACK,
                    )
                    return self.find_artifacts(
                        RequestTarget.previous_release(BRANCH_FALLBACK),
                        requested_variant,
                        target,
                        ignore_failed_push,
                        False,
                    )
                raise

            if artifact_urls is None:
                # This may be a new branch that hasn't had any artifacts built yet, so for that
                # case, we will fallback to searching the master branch for usable artifacts.
                self.handle_new_branch_fallback(
                    request.identifier, requested_variant, target, ignore_failed_push
                )

        elif request.request_type == RequestType.MONGO_PATCH_VERSION:
            tag = f"r{request.identifier}"
            commit = self.git_service.get_commit_from_tag(tag)
            artifact_urls = self.find_artifact_urls_for_commit(
                commit, requested_variant, target, ignore_failed_push
            )

        elif request.request_type == RequestType.GIT_COMMIT:
            artifact_urls = self.find_artifact_urls_for_commit(
                request.identifier, requested_variant, target, ignore_failed_push
            )

        else:
            raise ValueError(f"Unknown RequestType: '{request.request_type}'")

        return artifact_urls

    def handle_new_branch_fallback(
        self,
        mongo_version: str,
        requested_variant: Optional[str],
        target: DownloadTarget,
        ignore_failed_push: bool,
    ) -> Optional[EvgURLInfo]:
        """
        Handle the case a new branch does not have any built artifacts.

        In the case we are looking for artifacts on a new branch that has yet to build any
        artifacts, we will fallback to the master branch and look for the latest artifacts
        built before the branch split off.

        :param mongo_version: Version of mongo to target.
        :param requested_variant: Name of build variant that was requested.
        :param target: Attributes of the build variant to download from.
        :param ignore_failed_push: Whether a failed push should raise an exception.
        :return: Artifact URLs to use for new branch.
        """
        merge_base = self.git_service.get_merge_base_commit(mongo_version)
        if merge_base is None:
            LOGGER.warning("Unable to find merge-base commit", mongo_version=mongo_version)
            return None

        request = RequestTarget(request_type=RequestType.GIT_BRANCH, identifier=BRANCH_FALLBACK)
        return self.find_artifacts(
            request,
            requested_variant,
            target,
            ignore_failed_push,
            fallback_to_master=False,
            starting_commit=merge_base,
        )

    def find_artifact_urls_for_version(
        self,
        version_id: str,
        buildvariant_name: Optional[str],
        target: DownloadTarget,
        ignore_failed_push: bool,
    ) -> Optional[EvgURLInfo]:
        """
        Find URLs for downloading artifacts from the given evergreen version.

        :param version_id: Version ID of evergreen version to target.
        :param buildvariant_name: Name of build variant to download from.
        :param target: Details about what platform/edition/architecture to target.
        :param ignore_failed_push: Can a failed push task be ignored.
        :return: Collection of URLs to download artifacts.
        """
        evg_version = self.evg_service.get_version(version_id)
        if evg_version is None:
            return None

        if not buildvariant_name:
            buildvariant_name = self.infer_build_variant(evg_version, target)
            if buildvariant_name is None:
                raise ValueError(f"Unable to find matching build variant for '{version_id}'.")

        if buildvariant_name not in evg_version.build_variants_map:
            raise MissingBuildVariantError(
                f"Buildvariant '{buildvariant_name}' not found in evergreen. "
                f"Available buildvariants can be found in {SETUP_REPRO_ENV_CONFIG_FILE}."
            )

        urls = self.evg_service.get_compile_artifact_urls(
            evg_version, buildvariant_name, ignore_failed_push=ignore_failed_push
        )

        if urls is not None:
            return EvgURLInfo(
                urls=urls.urls,
                evg_version_id=evg_version.version_id,
                project_identifier=evg_version.project_identifier,
            )
        return None

    def scan_for_artifact_urls(
        self,
        mongo_project: str,
        requested_variant: Optional[str],
        target: DownloadTarget,
        ignore_failed_push: bool,
        starting_commit: Optional[str] = None,
    ) -> Optional[EvgURLInfo]:
        """
        Scan through evergreen version to find the most recent artifacts to download.

        :param mongo_project: Name of mongo project in evergreen.
        :param buildvariant_name: Name of build variant to download from.
        :param target: Details about what platform/edition/architecture to target.
        :param ignore_failed_push: Can a failed push task be ignored.
        :param starting_commit: Only match versions previous to this commit.
        :return: URLs for artifacts in most recent version found.
        """
        found_starting_commit = False
        for evg_version in self.evg_service.get_version_iterator(mongo_project):
            # If a starting commit was provided, only search after that commit.
            if starting_commit is not None and found_starting_commit is False:
                if evg_version.revision != starting_commit:
                    continue
                found_starting_commit = True

            # Skip all versions until we get the revision we should start looking from
            urls = self.find_artifact_urls_for_version(
                evg_version.version_id,
                requested_variant,
                target,
                ignore_failed_push=ignore_failed_push,
            )
            if urls is not None:
                return urls

            LOGGER.debug("Unable to find URLs for version", version=evg_version.version_id)

        return None

    def infer_build_variant(self, evg_version: Version, target: DownloadTarget) -> Optional[str]:
        """
        Determine what build variant to target.

        :param evg_version: Evergreen version to download from.
        :param target: Details of what download to target.
        :return: Name of build variant to download from.
        """
        evg_project = evg_version.project_identifier
        LOGGER.debug("Found evergreen project", evergreen_project=evg_project)

        version = self.project_to_version(evg_project)
        if version is None:
            raise ValueError(f"Unable to determine version of project: '{evg_project}'")

        if not target.is_complete():
            target = self.get_download_target(version, target)

        LOGGER.debug("Searching for build variant", version=version, target=target)
        buildvariant_name = self.setup_repro_config.find_matching_build_variant(target, version)
        LOGGER.debug("Found buildvariant", buildvariant_name=buildvariant_name)
        return buildvariant_name

    def get_download_target(self, mongo_version: str, target: DownloadTarget) -> DownloadTarget:
        """
        Determine the download target that should be used.

        This will infer the platform if it is not specified.

        :param mongo_version: Mongo version being targetted.
        :param target: Download target to use as a base.
        :return: Download target to use.
        """
        target_platform = target.platform
        if target_platform is None:
            target_platform = self.platform_service.infer_platform(target.edition, mongo_version)

        return DownloadTarget(
            edition=target.edition, platform=target_platform, architecture=target.architecture
        )

    @staticmethod
    def project_to_version(project_identifier: str) -> Optional[str]:
        """
        Determine the given version based on the given project identifier.

        :param project_identifier: Project identifier to use.
        :return: Version of given project identifier.
        """
        if project_identifier in NON_VERSION_PROJECTS:
            return project_identifier.replace(PROJECT_PREFIX, "")

        version_match = VERSION_PROJECT_RE.match(project_identifier)
        if version_match is not None:
            return version_match.group(1)

        return None

    def find_artifact_urls_for_commit(
        self,
        commit: str,
        requested_variant: Optional[str],
        target: DownloadTarget,
        ignore_failed_push: bool,
    ) -> Optional[EvgURLInfo]:
        """
        Find URLs for downloading artifacts from the given commit hash.

        :param commit: Git commit hash to search for.
        :param requested_variant: Name of build variant to download from.
        :param target: Details about what platform/edition/architecture to target.
        :param ignore_failed_push: Can a failed push task be ignored.
        :return: Collection of URLs to download artifacts.
        """
        versions = self.find_versions_for_commit(commit)

        if len(versions) == 0:
            raise ValueError(f"Unable to find commit '{commit}' in any evergreen projects")

        urls = None
        for version in versions:
            try:
                urls = self.find_artifact_urls_for_version(
                    version, requested_variant, target, ignore_failed_push
                )
            except MissingBuildVariantError:
                continue
            else:
                break

        return urls

    def find_versions_for_commit(self, commit_hash: str) -> List[str]:
        """
        Given a commit hash, find the evergreen versions that are associated with it.

        :param commit_hash: Git commit hash to search for.
        :return: List of evergreen version ids associated with the given git commit hash.
        """
        LOGGER.debug("Search evergreen projects for commit", commit_hash=commit_hash)
        evergreen_projects = self.evg_service.get_mongo_projects()
        found_versions = []
        for project in evergreen_projects:
            possible_version_id = f"{project}_{commit_hash}".replace("-", "_")
            LOGGER.debug(
                "Checking project for commit",
                project=project,
                possible_version_id=possible_version_id,
            )
            if self.evg_service.query_version_existence(possible_version_id):
                found_versions.append(possible_version_id)

        return found_versions


def find_fallback_project(searched_project: str) -> str:
    """
    Find the other project to try when searching on the master branch.

    :param searched_project: Project that has already been searched.
    :return: Other project to search.
    """
    other_project = list(NON_VERSION_PROJECTS.difference({searched_project}))[0]
    LOGGER.info(
        "Could not find expected build variant trying other project",
        searched_project=searched_project,
        other_project=other_project,
    )
    return other_project



"""Filesystem-related helper functions."""

import os
import tempfile
from typing import Any, Dict, Optional

import yaml


def mkdtemp_in_build_dir():
    """Use build/ as the temp directory since it's mapped to an EBS volume on Evergreen hosts."""
    build_tmp_dir = os.path.join("build", "tmp")
    os.makedirs(build_tmp_dir, exist_ok=True)

    return tempfile.mkdtemp(dir=build_tmp_dir)


def build_bin_path(parent: str) -> Optional[str]:
    """Find the first bin directory recursively in the parent directory."""
    for path, dirs, _ in os.walk(parent):
        if "bin" in dirs:
            return os.path.join(path, "bin")
    return None


def read_yaml_file(path: str) -> Dict[str, Any]:
    """
    Read the yaml file at the given path and return the contents.

    :param path: Path to file to read.
    :return: Contents of given file.
    """
    with open(path) as file_handle:
        return yaml.safe_load(file_handle)



"""Options used for building process."""
import re


class PathOptions:
    """Get options/methods to resolve path-related stuff."""

    main_binary_folder_name = "bin"
    shared_library_folder_name = "lib"

    # extend the below list if there are new types of shared libraries
    # below patterns should match (but not limited to):
    #  filename.so, filename.so.12, filename.so.debug
    #  filename.dylib
    _shared_library_file_patterns = [r".+\.so(\.(\d{1,}|debug))?$", r".+\.dylib$"]
    _compiled_shared_library_file_patterns = None

    @property
    def shared_library_file_patterns(self) -> tuple:
        """Cache & return compiled regex patterns for further usage."""

        if not self._compiled_shared_library_file_patterns:
            self._compiled_shared_library_file_patterns = tuple(
                map(re.compile, self._shared_library_file_patterns)
            )
        return self._compiled_shared_library_file_patterns or ()

    def is_shared_library_file(self, path: str) -> bool:
        """Check if given path points to a shared library file."""

        for pattern in self.shared_library_file_patterns:
            if pattern.match(path):
                return True
        return False

    def get_binary_folder_name(self, path: str) -> str:
        """Analyze path/file and return name of folder which contains that file."""

        if self.is_shared_library_file(path):
            return self.shared_library_folder_name
        # fallback to default, main binary folder
        return self.main_binary_folder_name



"""Service to parse backtraces in mongo log files."""
from json import JSONDecodeError, JSONDecoder
from typing import Any, Dict, List, NamedTuple, Optional

import inject

BACKTRACE_KEY = "backtrace"
PROCESS_INFO_KEY = "processInfo"

SOMAP_KEY = "somap"
MONGODB_VERSION_KEY = "mongodbVersion"

BINARY_LOAD_ADDRESS_KEY = "b"
ELF_TYPE_KEY = "elfType"
BUILD_ID_KEY = "buildId"
VMADDR_KEY = "vmaddr"


class SoInfo(NamedTuple):
    """Information about process that printed backtrace."""

    elf_type: Optional[int] = None
    vmaddr: Optional[str] = None
    build_id: Optional[str] = None
    mongodb_version: Optional[str] = None


class TraceDoc(NamedTuple):
    """
    Document with backtrace related information.

    * backtrace: Backtrace doc printed out by mongo process.
    * base_addr_map: Map from binary load address to description of library.
    """

    backtrace: List[Dict[str, Any]]
    base_addr_map: Dict[str, SoInfo]


class MongoLogParserService:
    """Service to parse backtraces in mongo log files."""

    @inject.autoparams()
    def __init__(self, json_decoder: JSONDecoder):
        """
        Initialize.

        :param json_decoder: JSONDecoder object.
        """
        self.json_decoder = json_decoder

    def parse_log_lines(
        self, log_lines: str, base_addr_map: Optional[Dict[str, SoInfo]] = None
    ) -> List[TraceDoc]:
        """
        Analyze log lines and return all backtraces found.

        :param log_lines: Log lines as string.
        :param base_addr_map: Map from binary load address to description of library.
        :return: List of trace docs.
        """
        global_base_addr_map = base_addr_map if base_addr_map is not None else {}
        backtraces_list = []

        for line in log_lines.splitlines():
            log_json = self.decode_log(line)
            if log_json is None:
                continue

            process_info = self.get_value_recursively(log_json, PROCESS_INFO_KEY)
            if process_info is not None:
                process_base_addr_map = self.make_base_addr_map(process_info)
                global_base_addr_map.update(process_base_addr_map)

            backtrace = self.get_value_recursively(log_json, BACKTRACE_KEY)
            if backtrace is not None:
                backtraces_list.append(backtrace)

        return [TraceDoc(backtrace, global_base_addr_map) for backtrace in backtraces_list]

    def decode_log(self, line: str) -> Optional[Dict[str, Any]]:
        """
        Decode JSON doc from log string.

        :param line: Log string.
        :return: Decoded log line or None.
        """
        start_index = line.find("{")
        if start_index == -1:
            return None

        # Skip over everything before the first '{' since it is likely to be log line prefixes.
        sub_line = line[start_index:]
        try:
            # Using raw_decode() to ignore extra data after the closing '}' to allow maximal
            # sloppiness in copy-pasting input.
            return self.json_decoder.raw_decode(sub_line)[0]

        except JSONDecodeError:
            # It is expected that any kind of logs may be passed in along with a backtrace,
            # we are skipping the lines that do not contain a valid JSON document
            return None

    def get_value_recursively(self, doc: Any, key: str) -> Optional[Any]:
        """
        Search the dict recursively for a value that has the key.

        :param doc: Dict or any value to search in.
        :param key: Key to search for.
        :return: Value of the key or None.
        """
        try:
            if key in doc.keys():
                return doc[key]
            for sub_doc in doc.values():
                res = self.get_value_recursively(sub_doc, key)
                if res:
                    return res
        except AttributeError:
            pass
        return None

    @staticmethod
    def make_base_addr_map(process_info: Dict[str, Any]) -> Dict[str, SoInfo]:
        """
        Make map from binary load address to description of library from the processInfo.

        :param process_info: Backtrace processInfo.
        :return: Map from binary load address to description of library.
        """
        addr_map = {}
        mongodb_version = process_info.get(MONGODB_VERSION_KEY)
        so_map_list = process_info.get(SOMAP_KEY, [])

        for so_entry in so_map_list:
            if BINARY_LOAD_ADDRESS_KEY in so_entry:
                addr_map[so_entry[BINARY_LOAD_ADDRESS_KEY]] = SoInfo(
                    elf_type=so_entry.get(ELF_TYPE_KEY),
                    vmaddr=so_entry.get(VMADDR_KEY),
                    build_id=so_entry.get(BUILD_ID_KEY),
                    mongodb_version=mongodb_version,
                )

        return addr_map



"""CLI for symbolize."""
import sys
from typing import List, Optional

import click
import inject

from db_contrib_tool.symbolizer.mongosymb import (
    InputFormat,
    OutputFormat,
    PathResolver,
    SymbolizerOrchestrator,
    SymbolizerParameters,
)
from db_contrib_tool.usage_analytics import CommandWithUsageTracking

INPUT_FORMATS = ("classic", "thin")
DEFAULT_INPUT_FORMAT = "classic"
OUTPUT_FORMATS = ("classic", "json")
DEFAULT_OUTPUT_FORMAT = "classic"


@click.command(cls=CommandWithUsageTracking)
@click.option("--symbolizer-path", type=click.Path(), help="Path to llvm-symbolizer executable.")
@click.option(
    "--dsym-hint", multiple=True, help="`-dsym-hint` flag values to pass to llvm-symbolizer."
)
@click.option(
    "--input-format",
    type=click.Choice(INPUT_FORMATS, case_sensitive=False),
    default=DEFAULT_INPUT_FORMAT,
    help="Input mongo log format type.",
)
@click.option(
    "--output-format",
    type=click.Choice(OUTPUT_FORMATS, case_sensitive=False),
    default=DEFAULT_OUTPUT_FORMAT,
    help="`json` shows some extra information.",
)
@click.option("--live", is_flag=True, help="Enter live mode.")
@click.option("--host", help="URL of web service running the API to get debug symbol URL.")
# caching mechanism is currently not fully developed and needs more advanced cleaning techniques,
# we add an option to enable it after completing the implementation.
@click.option(
    "--cache-dir",
    type=click.Path(),
    help="Full path to a directory to store cache/files.",
    hidden=True,
)
@click.option(
    "--total-seconds-for-retries", default=0, help="Timeout for getting data from web service."
)
@click.option("--client-secret", help="Secret key for Okta Oauth.")
@click.option("--client-id", help="Client id for Okta Oauth.")
def symbolize(
    symbolizer_path: Optional[str],
    dsym_hint: List[str],
    input_format: str,
    output_format: str,
    live: bool,
    host: Optional[str],
    cache_dir: Optional[str],
    total_seconds_for_retries: int,
    client_secret: Optional[str],
    client_id: Optional[str],
):
    """
    Symbolize MongoDB stack traces.

    To use as a script, paste the JSON object on the line after ----- BEGIN BACKTRACE ----- into the
    standard input of this script.

    Sample usage:

    db-contrib-tool symbolize --symbolizer-path=/path/to/llvm-symbolizer < /file/with/stacktrace

    or more simply:

    db-contrib-tool symbolize < file/with/stackraces

    You can also pass --output-format=json, to get rich json output. It shows some extra information,
    but emits json instead of plain text.

    Notes:

        - To symbolize stacktraces from Evergreen tasks, please ensure the
    `generate_buildid_to_debug_symbols_mapping` task is run on the build variant that generates
    the stacktrace to symbolize.

        - In a non-GUI environment, please forward local port 8989 to enable the OAuth flow from the
    local browser. The command may look something like:

            `ssh -L 8989:localhost:8989 user@virtual-host.amazonaws.com`.
    """

    symbolizer_params = SymbolizerParameters(
        symbolizer_path=symbolizer_path,
        dsym_hint=dsym_hint,
        input_format=InputFormat.CLASSIC if input_format == "classic" else InputFormat.THIN,
        output_format=OutputFormat.CLASSIC if output_format == "classic" else OutputFormat.JSON,
        live=live,
        host=host,
        cache_dir=cache_dir,
        total_seconds_for_retries=total_seconds_for_retries,
        client_secret=client_secret,
        client_id=client_id,
    )

    dbg_path_resolver = PathResolver(
        host=symbolizer_params.host,
        cache_dir=symbolizer_params.cache_dir,
        client_secret=symbolizer_params.client_secret,
        client_id=symbolizer_params.client_id,
    )

    def dependencies(binder: inject.Binder) -> None:
        """Define dependencies for execution."""
        binder.bind(PathResolver, dbg_path_resolver)

    inject.configure(dependencies)
    symbolize_orchestrator = inject.instance(SymbolizerOrchestrator)

    success = symbolize_orchestrator.execute(symbolizer_params)
    if not success:
        sys.exit(1)



"""Client for working with git."""
import subprocess
from typing import List, NamedTuple, Optional

import structlog

LOGGER = structlog.get_logger(__name__)

DEFAULT_ENCODING = "utf-8"


class GitResult(NamedTuple):
    """
    Results from running a git command.

    * return_code: Exit code of the run command.
    * output: Stdout of the run command.
    * error: Stderr of the run command.
    """

    return_code: int
    output: str
    error: str


class GitClient:
    """A client for interacting with git."""

    def merge_base(self, remote: str, branch: str) -> GitResult:
        """
        Run the merge-base command.

        :param remote: Remote repository to run against.
        :param branch: branch to pass to merge-base.
        :return: The results of running the command.
        """
        args = ["git", "merge-base", remote, branch]
        return self._run_command(args)

    def rev_parse(self, commit: str, abbrev_ref: bool = False, verify: bool = False) -> GitResult:
        """
        Run the rev-parse command.

        :param commit: Commit to run against
        :param abbrev_ref: Include the --abbrev-ref flag.
        :param verify: Include the --verify flag.
        :return: The results of running the command.
        """
        args = ["git", "rev-parse"]

        if abbrev_ref:
            args.append("--abbrev-ref")

        if verify:
            args.append("--verify")

        args.append(commit)
        return self._run_command(args)

    def rev_list(self, commit: str, number: Optional[int]) -> GitResult:
        """
        Run the rev-list command.

        :param commit: Commit to run against.
        :param number: The number of commits to output.
        :return: The results of running the command.
        """
        args = ["git", "rev-list"]

        if number:
            args.append(f"-{number}")

        args.append(commit)
        return self._run_command(args)

    @staticmethod
    def _run_command(args: List[str]) -> GitResult:
        """
        Run the given command and return the results.

        :param args: Command with arguments to run.
        :return: The results of running the command.
        """
        result = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False)
        output = result.stdout.decode(DEFAULT_ENCODING).strip()
        error = result.stderr.decode(DEFAULT_ENCODING).strip()
        LOGGER.debug("Git command results", cmd=args, output=output, error=error)
        return GitResult(
            return_code=result.returncode,
            output=output,
            error=error,
        )



"""Helper functions to interact with evergreen."""
import os
import pathlib
from typing import List

import structlog
from evergreen import RetryingEvergreenApi

EVERGREEN_HOST = "https://evergreen.mongodb.com"
EVERGREEN_CONFIG_LOCATIONS = (
    # Common for machines in Evergreen
    os.path.join(os.getcwd(), ".evergreen.yml"),
    # Common for local machines
    os.path.expanduser(os.path.join("~", ".evergreen.yml")),
)

LOGGER = structlog.getLogger(__name__)


class EvergreenConnError(Exception):
    """Errors in evergreen_conn.py."""

    pass


def _find_evergreen_yaml_candidates() -> List[str]:
    # Common for machines in Evergreen
    candidates = [os.getcwd()]

    cwd = pathlib.Path(os.getcwd())
    # add every path that is the parent of CWD as well
    for parent in cwd.parents:
        candidates.append(parent)

    # Common for local machines
    candidates.append(os.path.expanduser(os.path.join("~", ".evergreen.yml")))

    out = []
    for path in candidates:
        file = os.path.join(path, ".evergreen.yml")
        if os.path.isfile(file):
            out.append(file)

    return out


def get_evergreen_api(evergreen_config=None):
    """Return evergreen API."""
    if evergreen_config:
        possible_configs = [evergreen_config]
    else:
        possible_configs = _find_evergreen_yaml_candidates()

    if not possible_configs:
        LOGGER.error("Could not find .evergreen.yml", candidates=possible_configs)
        raise RuntimeError("Could not find .evergreen.yml")

    last_ex = None
    for config in possible_configs:
        try:
            return RetryingEvergreenApi.get_api(config_file=config)
        except Exception as ex:
            last_ex = ex
            continue

    LOGGER.error(
        "Could not connect to Evergreen with any .evergreen.yml files available on this system",
        config_file_candidates=possible_configs,
    )
    raise last_ex



"""Script for symbolizing MongoDB stack traces."""

import json
import os
import signal
import subprocess
import sys
import time
from collections import OrderedDict
from enum import Enum
from typing import IO, Any, AnyStr, Dict, List, NamedTuple, Optional

import inject
import requests
from tenacity import Retrying, retry_if_result, stop_after_delay, wait_fixed

from db_contrib_tool.clients.io_client import IOClient
from db_contrib_tool.symbolizer.mongo_log_parser_service import (
    BINARY_LOAD_ADDRESS_KEY,
    BUILD_ID_KEY,
    MongoLogParserService,
    SoInfo,
    TraceDoc,
)
from db_contrib_tool.utils.build_system_options import PathOptions
from db_contrib_tool.utils.oauth import (
    Configs,
    get_client_cred_oauth_credentials,
    get_oauth_credentials,
)

SYMBOLIZER_PATH_ENV = "MONGOSYMB_SYMBOLIZER_PATH"
DEFAULT_SYMBOLIZER_PATH = "/opt/mongodbtoolchain/v4/bin/llvm-symbolizer"


class CachedResults(object):
    """
    Used to manage / store results in a cache form (using dict as an underlying data structure).

    Idea is to allow only N items to be present in cache at a time and eliminate extra items on the go.
    """

    def __init__(self, max_cache_size: int, initial_cache: Dict[str, str] = None):
        """
        Initialize instance.

        :param max_cache_size: max number of items that can be added to cache
        :param initial_cache: initial items as dict
        """
        self._max_cache_size = max_cache_size
        self._cached_results = OrderedDict(initial_cache or {})

    def insert(self, key: str, value: str) -> Dict[str, str] or None:
        """
        Insert new data into cache.

        :param key: key string
        :param value: value string
        :return: inserted data as dict or None (if not possible to insert)
        """
        if self._max_cache_size <= 0:
            # we can't insert into 0-length dict
            return None

        if len(self._cached_results) >= self._max_cache_size:
            # remove items causing the size overflow of cache
            # we use FIFO order when removing objects from cache,
            # so that we delete olds and keep track of only the recent ones
            keys_iterator = iter(self._cached_results.keys())
            while len(self._cached_results) >= self._max_cache_size:
                # pop the first (the oldest) item in dict
                self._cached_results.pop(next(keys_iterator))

        if key not in self._cached_results:
            # actual insert operation
            self._cached_results[key] = value

        return dict(build_id=value)

    def get(self, key: str) -> str or None:
        """
        Try to get object by key.

        :param key: key string
        :return: value for key
        """
        if self._max_cache_size <= 0:
            return None

        return self._cached_results.get(key)


class PathResolver(object):
    """
    Class to find path for given buildId.

    We'll be sending request each time to another server to get path.
    This process is fairly small, but can be heavy in case of increased amount of requests per second.
    Thus, I'm implementing a caching mechanism (as a suggestion).
    It keeps track of the last N results from server, we always try to search from that cache, if not found then send
    request to server and cache the response for further usage.
    Cache size differs according to the situation, system resources and overall decision of development team.
    """

    # the main (API) sever that we'll be sending requests to
    default_host = "https://symbolizer-service.server-tig.prod.corp.mongodb.com"
    default_cache_dir = os.path.join(os.getcwd(), "build", "symbolizer_downloads_cache")
    default_creds_file_path = os.path.join(os.getcwd(), ".symbolizer_credentials.json")
    default_client_credentials_scope = "servertig-symbolizer-fullaccess"
    default_client_credentials_user_name = "client-user"

    def __init__(
        self,
        host: str = None,
        cache_size: int = 0,
        cache_dir: str = None,
        client_credentials_scope: str = None,
        client_credentials_user_name: str = None,
        client_id: str = None,
        client_secret: str = None,
        redirect_port: int = None,
        scope: str = None,
        auth_domain: str = None,
    ):
        """
        Initialize instance.

        :param host: URL of host - web service
        :param cache_size: size of cache. We try to cache recent results and use them instead of asking from server.
        Use 0 (by default) to disable caching
        """
        self.host = host or self.default_host
        self._cached_results = CachedResults(max_cache_size=cache_size)
        self.cache_dir = cache_dir or self.default_cache_dir
        self.client_credentials_scope = (
            client_credentials_scope or self.default_client_credentials_scope
        )
        self.client_credentials_user_name = (
            client_credentials_user_name or self.default_client_credentials_user_name
        )
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_port = redirect_port
        self.scope = scope
        self.auth_domain = auth_domain
        self.configs = Configs(
            client_credentials_scope=self.client_credentials_scope,
            client_credentials_user_name=self.client_credentials_user_name,
            client_id=self.client_id,
            auth_domain=self.auth_domain,
            redirect_port=self.redirect_port,
            scope=self.scope,
        )
        self.http_client = requests.Session()
        self.path_options = PathOptions()

        # create cache dir if it doesn't exist
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        self.authenticate()

    def authenticate(self):
        """Login & get credentials for further requests to web service."""

        # try to read from file
        if os.path.exists(self.default_creds_file_path):
            with open(self.default_creds_file_path) as cfile:
                data = json.loads(cfile.read())
                access_token, expire_time = data.get("access_token"), data.get("expire_time")
                if time.time() < expire_time:
                    # credentials not expired yet
                    self.http_client.headers.update({"Authorization": f"Bearer {access_token}"})
                    return

        if self.client_id and self.client_secret:
            # auth using secrets
            credentials = get_client_cred_oauth_credentials(
                self.client_id, self.client_secret, self.configs
            )
        else:
            # since we don't have access to secrets, ask user to auth manually
            credentials = get_oauth_credentials(configs=self.configs, print_auth_url=True)

        self.http_client.headers.update({"Authorization": f"Bearer {credentials.access_token}"})

        # write credentials to local file for further usage
        with open(self.default_creds_file_path, "w") as cfile:
            cfile.write(
                json.dumps(
                    {
                        "access_token": credentials.access_token,
                        "expire_time": time.time() + credentials.expires_in,
                    }
                )
            )

    @staticmethod
    def is_valid_path(path: str) -> bool:
        """
        Sometimes the given path may not be valid: e.g: path for a non-existing file.

        If we need to do extra checks on path, we'll do all of them here.
        :param path: path string
        :return: bool indicating the validation status
        """
        return os.path.exists(path)

    def get_from_cache(self, key: str) -> str or None:
        """
        Try to get value from cache.

        :param key: key string
        :return: value or None (if doesn't exist)
        """
        return self._cached_results.get(key)

    def add_to_cache(self, key: str, value: str) -> Dict[str, str]:
        """
        Add new value to cache.

        :param key: key string
        :param value: value string
        :return: added data as dict
        """
        return self._cached_results.insert(key, value)

    @staticmethod
    def url_to_filename(url: str) -> str:
        """
        Convert URL to local filename.

        :param url: download URL
        :return: full name for local file
        """
        return url.split("/")[-1]

    @staticmethod
    def unpack(path: str) -> str:
        """
        Use to utar/unzip files.

        :param path: full path of file
        :return: full path of 'bin' directory of unpacked file
        """
        out_dir = path.replace(".tgz", "", 1)
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        args = ["tar", "xopf", path, "-C", out_dir, "--strip-components 1"]
        cmd = " ".join(args)
        subprocess.check_call(cmd, shell=True)

        return out_dir

    def download(self, url: str) -> (str, bool):
        """
        Use to download file from URL.

        :param url: URL string
        :return: full path of downloaded file in local filesystem, bool indicating if file is already downloaded or not
        """
        exists_locally = False
        filename = self.url_to_filename(url)
        path = os.path.join(self.cache_dir, filename)
        if not os.path.exists(path):
            print("Downloading the file...")
            with requests.get(url, stream=True) as response:
                with open(path, "wb") as file:
                    for chunk in response.iter_content(chunk_size=2 * 1024 * 1024):
                        file.write(chunk)
        else:
            print("File already exists in cache")
            exists_locally = True
        return path, exists_locally

    def get_dbg_file(self, soinfo: SoInfo) -> Optional[str]:
        """
        To get path for given buildId.

        :param soinfo: Information about process that printed backtrace.
        :return: Path as string or None (if path not found)
        """
        build_id = soinfo.build_id.lower() if soinfo.build_id is not None else ""
        version = soinfo.mongodb_version
        binary_name = "mongo"
        # search from cached results
        path = self.get_from_cache(build_id)
        if not path:
            # path does not exist in cache, so we send request to server
            try:
                search_parameters = {"build_id": build_id}
                if version:
                    search_parameters["version"] = version
                print(f"Getting data from service... Search parameters: {search_parameters}")
                response = self.http_client.get(f"{self.host}/find_by_id", params=search_parameters)
                if response.status_code != 200:
                    sys.stderr.write(
                        f"Server returned unsuccessful status: {response.status_code}, "
                        f"response body: {response.text}\n"
                    )
                    return None
                else:
                    data = response.json().get("data", {})
                    path, binary_name = data.get("debug_symbols_url"), data.get("file_name")
            except Exception as err:  # noqa
                sys.stderr.write(
                    f"Error occurred while trying to get response from server "
                    f"for buildId({build_id}): {err}\n"
                )
                return None

            # update cached results
            if path:
                self.add_to_cache(build_id, path)

        if not path:
            return None

        # download & unpack debug symbols file and assign `path` to unpacked file's local path
        try:
            dl_path, exists_locally = self.download(path)
            if exists_locally:
                path = dl_path.replace(".tgz", "", 1)
            else:
                sys.stdout.write("Downloaded, now unpacking...\n")
                path = self.unpack(dl_path)
        except Exception as err:  # noqa
            sys.stderr.write(f"Failed to download & unpack file: {err}\n")
        # we may have '<name>.debug', '<name>.so' or just executable binary file which may not have file 'extension'.
        # if file has extension, it is good. if not, we should append .debug, because those without extension are
        # from release builds, and their debug symbol files contain .debug extension.
        # we need to map those 2 different file names ('<name>' becomes '<name>.debug').
        if not binary_name.endswith(".debug"):
            binary_name = f"{binary_name}.debug"

        inner_folder_name = self.path_options.get_binary_folder_name(binary_name)

        return os.path.join(path, inner_folder_name, binary_name)


class InputFormat(str, Enum):
    """Input mongo log format types."""

    CLASSIC = "classic"
    THIN = "thin"


class OutputFormat(str, Enum):
    """Output format types."""

    CLASSIC = "classic"
    JSON = "json"


class SymbolizerParameters(NamedTuple):
    """
    Parameters describing how logs should be symbolized.

    * symbolizer_path: Symbolizer executable path.
    * dsym_hint: List of `-dsym-hint` flag values to pass to symbolizer.

    * input_format: Input mongo log format.
    * output_format: Output format.

    * live: Whether it should enter live mode.

    * host: URL of web service running the API to get debug symbol URL.
    * cache_dir: Full path to a directory to store cache/files.
    * total_seconds_for_retries: Timeout for getting data from web service.

    * client_secret: Secret key for Okta Oauth.
    * client_id: Client id for Okta Oauth.
    """

    symbolizer_path: str
    dsym_hint: List[str]

    input_format: InputFormat
    output_format: OutputFormat

    live: bool

    host: str
    cache_dir: str
    total_seconds_for_retries: int

    client_secret: str
    client_id: str


class SymbolizerOrchestrator:
    """Orchestrator for log symbolizer."""

    @inject.autoparams()
    def __init__(
        self,
        mongo_log_parser_service: MongoLogParserService,
        io_client: IOClient,
        dbg_path_resolver: PathResolver,
    ) -> None:
        """
        Initialize.

        :param mongo_log_parser_service: Service to parse backtraces in log files.
        :param io_client: Client for working with input/output.
        :param dbg_path_resolver: Debug symbols file path resolver.
        """
        self.mongo_log_parser_service = mongo_log_parser_service
        self.io_client = io_client
        self.dbg_path_resolver = dbg_path_resolver

    def substitute_stdin(self, params: SymbolizerParameters) -> None:
        """
        Accept stdin stream as source of logs and symbolize it.

        :param params: Symbolizer parameters.
        """

        # Ignore Ctrl-C. When the process feeding the pipe exits, `stdin` will be closed.
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        print("Live mode activated, waiting for input...")
        global_base_addr_map = {}
        while True:
            backtrace_indicator = '{"backtrace":'
            line = sys.stdin.readline()
            if not line:
                return

            line = line.strip()

            if "Frame: 0x" in line:
                continue

            if backtrace_indicator in line:
                backtrace_index = line.index(backtrace_indicator)
                prefix = line[:backtrace_index]
                trace_docs = self.mongo_log_parser_service.parse_log_lines(
                    line, global_base_addr_map
                )
                if len(trace_docs) == 0:
                    print("Could not find json backtrace object in input...")
                    continue
                for trace_doc in trace_docs:
                    frames = self.symbolize_frames(trace_doc, params)
                    print(prefix)
                    print("Symbolizing...")
                    self.classic_output(frames, sys.stdout)
                    global_base_addr_map = trace_doc.base_addr_map
                print("Symbolization completed, waiting for input...")
            else:
                print(line)

    def execute(self, params: SymbolizerParameters) -> bool:
        """
        Execute symbolizer.

        :param params: Symbolizer parameters.
        :return: Succeeded or not.
        """
        if params.live:
            print("Entering live mode")
            self.substitute_stdin(params)
            return True

        log_lines = self.io_client.read_from_stdin()
        if not log_lines or not log_lines.strip():
            print(
                "Please provide the backtrace through stdin for symbolization;"
                "e.g. `your/symbolization/command < /file/with/stacktrace`"
            )

        trace_docs = self.mongo_log_parser_service.parse_log_lines(log_lines)
        if len(trace_docs) == 0:
            print("Could not find json backtrace object in input", file=sys.stderr)
            return False

        output_fn = None
        if params.output_format == OutputFormat.JSON:
            output_fn = json.dump
        if params.output_format == OutputFormat.CLASSIC:
            output_fn = self.classic_output

        for trace_doc in trace_docs:
            frames = self.symbolize_frames(trace_doc, params)
            output_fn(frames, sys.stdout, indent=2)

        return True

    def produce_frames(self, trace_doc: TraceDoc) -> List[Dict[str, Any]]:
        """
        Return a list of frame dicts from an object of trace doc.

        :param trace_doc: Traceback doc.
        :return: List of traceback frames.
        """

        frames = []
        for frame in trace_doc.backtrace:
            if BINARY_LOAD_ADDRESS_KEY not in frame:
                print(
                    f"Ignoring frame {frame} as it's missing the `{BINARY_LOAD_ADDRESS_KEY}` field;"
                    f" See SERVER-58863 for discussions"
                )
                continue
            soinfo = trace_doc.base_addr_map.get(frame[BINARY_LOAD_ADDRESS_KEY], SoInfo())
            if soinfo.elf_type == 3:
                addr_base = "0"
            elif soinfo.elf_type == 2:
                addr_base = frame[BINARY_LOAD_ADDRESS_KEY]
            else:
                addr_base = soinfo.vmaddr if soinfo.vmaddr is not None else "0"
            addr = int(addr_base, 16) + int(frame["o"], 16)
            # addr currently points to the return address which is the one *after* the call. x86 is
            # variable length so going backwards is difficult. However, llvm-symbolizer seems to do the
            # right thing if we just subtract 1 byte here. This has the downside of also adjusting the
            # address of instructions that cause signals (such as segfaults and divide-by-zero) which
            # are already correct, but there doesn't seem to be a reliable way to detect that case.
            addr -= 1
            frames.append(
                dict(
                    path=self.dbg_path_resolver.get_dbg_file(soinfo),
                    buildId=soinfo.build_id,
                    offset=frame["o"],
                    addr="0x{:x}".format(addr),
                    symbol=frame.get("s", None),
                )
            )
        return frames

    def symbolize_frames(
        self, trace_doc: TraceDoc, params: SymbolizerParameters
    ) -> List[Dict[str, Any]]:
        """
        Return a list of symbolized stack frames from a trace_doc in MongoDB stack dump format.

        :param trace_doc: Traceback doc.
        :param params: Symbolizer parameters.
        :return: List of traceback frames.
        """

        frames = self.preprocess_frames_with_retries(
            trace_doc,
            params.input_format,
            params.total_seconds_for_retries,
        )

        symbolizer_path = params.symbolizer_path
        if not symbolizer_path:
            symbolizer_path = os.environ.get(SYMBOLIZER_PATH_ENV)
            if not symbolizer_path:
                print(
                    f"Env value for '{SYMBOLIZER_PATH_ENV}' not found, using '{DEFAULT_SYMBOLIZER_PATH}' "
                    f"as a defualt executable path."
                )
                symbolizer_path = DEFAULT_SYMBOLIZER_PATH

        symbolizer_args = [symbolizer_path]
        for dh in params.dsym_hint:
            symbolizer_args.append("-dsym-hint={}".format(dh))
        symbolizer_process = subprocess.Popen(
            args=symbolizer_args,
            close_fds=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=sys.stdout,
        )

        for frame in frames:
            if frame["path"] is None:
                print("Path not found in frame:", frame)
                continue
            symbol_line = "CODE {path:} {addr:}\n".format(**frame)
            symbolizer_process.stdin.write(symbol_line.encode())
            symbolizer_process.stdin.flush()
            frame["symbinfo"] = self.extract_symbols(symbolizer_process.stdout)
        symbolizer_process.stdin.close()
        symbolizer_process.wait()
        return frames

    @staticmethod
    def extract_symbols(stdout: Optional[IO[AnyStr]]) -> List[Dict[str, Any]]:
        """
        Extract symbol information from the output of llvm-symbolizer.

        Return a list of dictionaries, each of which has fn, file, column and line entries.

        The format of llvm-symbolizer output is that for every CODE line of input,
        it outputs zero or more pairs of lines, and then a blank line. This way, if
        a CODE line of input maps to several inlined functions, you can use the blank
        line to find the end of the list of symbols corresponding to the CODE line.

        The first line of each pair contains the function name, and the second contains the file,
        column and line information.

        :param stdout: Output of llvm-symbolizer.
        :return: List of symbol information.
        """
        result = []
        step = 0
        while True:
            line = stdout.readline().decode()
            if line == "\n":
                break
            if step == 0:
                result.append({"fn": line.strip()})
                step = 1
            else:
                file_name, line, column = line.strip().rsplit(":", 3)
                result[-1].update({"file": file_name, "column": int(column), "line": int(line)})
                step = 0
        return result

    def preprocess_frames(
        self, trace_doc: TraceDoc, input_format: InputFormat
    ) -> List[Dict[str, Any]]:
        """
        Process the paths in frame objects.

        :param trace_doc: Traceback doc.
        :param input_format: Input format.
        :return: List of traceback frames.
        """

        if input_format == InputFormat.CLASSIC:
            return self.produce_frames(trace_doc)
        elif input_format == InputFormat.THIN:
            frames = trace_doc.backtrace
            for frame in frames:
                soinfo = SoInfo(build_id=frame.get(BUILD_ID_KEY))
                frame["path"] = self.dbg_path_resolver.get_dbg_file(soinfo)
            return frames

    @staticmethod
    def has_high_not_found_paths_ratio(frames: List[Dict[str, Any]]) -> bool:
        """
        Check whether not found paths in frames ratio is higher than 0.5.

        :param frames: the list of traceback frames
        :return: True if ratio is higher than 0.5
        """
        not_found = [1 for f in frames if f.get("path") is None]
        not_found_ratio = len(not_found) / (len(frames) or 1)
        return not_found_ratio >= 0.5

    def preprocess_frames_with_retries(
        self,
        trace_doc: TraceDoc,
        input_format: InputFormat,
        total_seconds_for_retries: int = 0,
    ) -> List[Dict[str, Any]]:
        """
        Process the paths in frame objects.

        :param trace_doc: Traceback doc.
        :param input_format: Input format.
        :param total_seconds_for_retries: Timeout for retries.
        :return: List of traceback frames.
        """

        retrying = Retrying(
            retry=retry_if_result(self.has_high_not_found_paths_ratio),
            wait=wait_fixed(60),
            stop=stop_after_delay(total_seconds_for_retries),
            retry_error_callback=lambda retry_state: retry_state.outcome.result(),
        )

        return retrying(self.preprocess_frames, trace_doc, input_format)

    @staticmethod
    def classic_output(frames, outfile, **kwargs):
        """Provide classic output."""
        for frame in frames:
            symbinfo = frame.get("symbinfo")
            if symbinfo:
                for sframe in symbinfo:
                    outfile.write(" {file:s}:{line:d}:{column:d}: {fn:s}\n".format(**sframe))
            else:
                outfile.write(
                    " Couldn't extract symbols: path={path}\n".format(
                        path=frame.get("path", "no value found")
                    )
                )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['db_contrib_tool',
 'db_contrib_tool.clients',
 'db_contrib_tool.evg_aware_bisect',
 'db_contrib_tool.services',
 'db_contrib_tool.setup_repro_env',
 'db_contrib_tool.symbolizer',
 'db_contrib_tool.utils']

package_data = \
{'': ['*'], 'db_contrib_tool': ['config/*']}

install_requires = \
['Inject>=4.3.1,<5.0.0',
 'PyGithub==1.55',
 'PyYAML==6.0',
 'analytics-python==1.4.0',
 'click>=8.1.3,<9.0.0',
 'distro==1.6.0',
 'evergreen.py==3.4.5',
 'oauthlib==3.1.1',
 'pkce==1.0.3',
 'pydantic==1.8.2',
 'requests-oauthlib==1.3.0',
 'requests==2.26.0',
 'structlog==21.4.0',
 'tenacity>=8.0.1,<9.0.0']

entry_points = \
{'console_scripts': ['db-contrib-tool = db_contrib_tool.cli:cli']}

setup_kwargs = {
    'name': 'db-contrib-tool',
    'version': '0.5.0',
    'description': "The `db-contrib-tool` - MongoDB's tool for contributors.",
    'long_description': '# db-contrib-tool\n\nThe `db-contrib-tool` - MongoDB\'s tools for contributors.\n\n## Table of contents\n\n- [db-contrib-tool](#db-contrib-tool)\n  - [Table of contents](#table-of-contents)\n  - [Description](#description)\n  - [Dependencies](#dependencies)\n  - [Installation](#installation)\n  - [Usage](#usage)\n  - [Contributor\'s Guide (local development)](#contributors-guide-local-development)\n    - [Install project dependencies](#install-project-dependencies)\n    - [Run command line tool (local development)](#run-command-line-tool-local-development)\n    - [Run linters](#run-linters)\n    - [Run tests](#run-tests)\n    - [Pre-commit](#pre-commit)\n    - [Testing changes in mongo](#testing-changes-in-mongo)\n    - [Test pipx package](#test-pipx-package)\n    - [Versioning](#versioning)\n    - [Code Review](#code-review)\n    - [Deployment](#deployment)\n\n## Description\n\nThe command line tool with various subcommands:\n- `bisect` - performs an evergreen-aware git-bisect to find the \'last passing version\' and \'first failing version\' of mongo\n- `setup-repro-env`\n  - [README.md](https://github.com/10gen/db-contrib-tool/blob/main/src/db_contrib_tool/setup_repro_env/README.md)\n  - downloads and installs:\n    - particular MongoDB versions\n    - debug symbols\n    - artifacts (including resmoke, python scripts etc)\n    - python venv for resmoke, python scripts etc\n- `symbolize`\n  - [README.md](https://github.com/10gen/db-contrib-tool/blob/main/src/db_contrib_tool/symbolizer/README.md)\n  - Symbolizes stacktraces from recent `mongod` and `mongos` binaries compiled in Evergreen, including patch builds, mainline builds, and release/production builds.\n  - Requires authenticating to an internal MongoDB symbol mapping service.\n\n## Dependencies\n\n- Python 3.7 or later (python3 from the [MongoDB Toolchain](https://github.com/10gen/toolchain-builder/blob/master/INSTALL.md) is highly recommended)\n\n## Installation\n\nMake sure [dependencies](#dependencies) are installed.\nUse [pipx](https://pypa.github.io/pipx/) to install db-contrib-tool that will be available globally on your machine:\n```bash\n$ python3 -m pip install pipx\n$ python3 -m pipx ensurepath\n```\n\nInstalling db-contrib-tool:\n```bash\n$ python3 -m pipx install db-contrib-tool\n```\n\nUpgrading db-contrib-tool:\n```bash\n$ python3 -m pipx upgrade db-contrib-tool\n```\n\n## Usage\n\nPrint out help message:\n```bash\n$ db-contrib-tool -h\n```\nMore information on the usage of `setup-repro-env` can be found [here](https://github.com/10gen/db-contrib-tool/blob/main/src/db_contrib_tool/setup_repro_env/README.md).\n\n## Contributor\'s Guide (local development)\n\n### Install project dependencies\n\nThis project uses [poetry](https://python-poetry.org/) for dependency management.\n```bash\n$ poetry install\n```\n\n### Run command line tool (local development)\n\n```bash\n$ ENV=DEV poetry run db-contrib-tool -h\n```\n\n### Run linters\n\n```bash\n$ poetry run isort src tests\n$ poetry run black src tests\n```\n\n### Run tests\n\n```bash\n$ poetry run pytest\n```\n\n### Pre-commit\n\nThis project has [pre-commit](https://pre-commit.com/) configured. Pre-commit will run\nconfigured checks at git commit time.<br>\nTo enable pre-commit on your local repository run:\n```bash\n$ poetry run pre-commit install\n```\n\nTo run pre-commit manually:\n```bash\n$ poetry run pre-commit run\n```\n\n### Testing changes in mongo\n\nThis tool is used to help run tests in the mongodb/mongo repository. On occasion, it may be\ndesirable to run a mongodb-mongo-* patch build with in-flight changes to this repository. The\nfollowing steps can be take to accomplish that.\n\n- Create a branch with the changes you wish to test.\n- Push the branch to the origin repository: `git push -u origin <branch_name>`.\n- In the "mongo" repository, edit the [evergreen/prelude_db_contrib_tool.sh](https://github.com/10gen/mongo/blob/750f2cabe28900e86950695b7711c860a4165c94/evergreen/prelude_db_contrib_tool.sh#L16)\n  to install from the git repository instead of from pypi:\n\n  ```bash\n  python -m pip --disable-pip-version-check install "git+ssh://git@github.com/10gen/db-contrib-tool.git@<branch_name>" || exit 1\n  ```\n\n- Create a patch build.\n\nThe patch build should now pull down the changes from your branch instead of using the published\ndb-contrib-tool.\n\n**Note**: Since the db-contrib-tool is pulled from your branch, if you need to make additional\nchanges to the tool, you can just push to the branch and then restart the desired tasks. There is\nno need to create an additional patch build unless you also need to make updates to the mongo\nrepository.\n\n### Test pipx package\n\nPipx installation recommendations can be found in [installation](#installation) section.<br>\nThe tool can be installed via pipx from your local repo:\n```bash\n$ python3 -m pipx install /path/to/db-contrib-tool\n```\n\n### Versioning\n\nThis project uses [semver](https://semver.org/) for versioning.\nPlease include a description what is added for each new version in `CHANGELOG.md`.\n\n### Code Review\n\nPlease open a Github Pull Request for code review.\nThis project uses the [Evergreen Commit Queue](https://github.com/evergreen-ci/evergreen/wiki/Commit-Queue#pr).\nAdd a PR comment with `evergreen merge` to trigger a merge.\n\n### Deployment\n\nDeployment to pypi is automatically triggered on merges to main.\n',
    'author': 'STM team',
    'author_email': 'dev-prod-stm@10gen.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/10gen/db-contrib-tool',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""CLI for setup_repro_env."""
import logging
import os
import sys
from typing import List, Optional

import click
import inject
import structlog
from click.core import ParameterSource
from evergreen import EvergreenApi

from db_contrib_tool.clients.resmoke_proxy import ResmokeProxy
from db_contrib_tool.config import (
    SETUP_REPRO_ENV_CONFIG,
    SETUP_REPRO_ENV_CONFIG_FILE,
    SetupReproEnvConfig,
)
from db_contrib_tool.setup_repro_env.download_service import DownloadOptions
from db_contrib_tool.setup_repro_env.setup_repro_env import (
    SetupReproOrchestrator,
    SetupReproParameters,
)
from db_contrib_tool.usage_analytics import CommandWithUsageTracking
from db_contrib_tool.utils.evergreen_conn import (
    EVERGREEN_CONFIG_LOCATIONS,
    get_evergreen_api,
)

DEFAULT_INSTALL_DIR = os.path.join(os.getcwd(), "build", "multiversion_bin")
DEFAULT_LINK_DIR = os.getcwd()
DEFAULT_WITH_ARTIFACTS_INSTALL_DIR = os.path.join(os.getcwd(), "repro_envs")
DEFAULT_WITH_ARTIFACTS_LINK_DIR = os.path.join(
    DEFAULT_WITH_ARTIFACTS_INSTALL_DIR, "multiversion_bin"
)

EDITIONS = ("base", "enterprise", "targeted")
DEFAULT_EDITION = "enterprise"
ARCHITECTURES = ("x86_64", "arm64", "aarch64", "ppc64le", "s390x")
DEFAULT_ARCHITECTURE = "x86_64"

DEFAULT_RESMOKE_CMD = "python buildscripts/resmoke.py"

EXTERNAL_LOGGERS = [
    "evergreen",
    "github",
    "inject",
    "segment",
    "urllib3",
]


def setup_logging(debug=False):
    """Enable logging."""
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(
        format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
        level=log_level,
        stream=sys.stdout,
    )
    for logger in EXTERNAL_LOGGERS:
        logging.getLogger(logger).setLevel(logging.WARNING)
    structlog.configure(logger_factory=structlog.stdlib.LoggerFactory())


@click.command(cls=CommandWithUsageTracking)
@click.pass_context
@click.argument(
    "versions",
    nargs=-1,
    required=False,
)
@click.option(
    "-i",
    "--installDir",
    "install_dir",
    type=click.Path(),
    default=DEFAULT_INSTALL_DIR,
    help="Directory to install the download archive.",
    show_default=f"`{DEFAULT_INSTALL_DIR}`, if --downloadArtifacts is passed"
    f" the value defaults to `{DEFAULT_WITH_ARTIFACTS_INSTALL_DIR}`",
)
@click.option(
    "-l",
    "--linkDir",
    "link_dir",
    type=click.Path(),
    default=DEFAULT_LINK_DIR,
    help=f"Directory to install the download archive.",
    show_default=f"`{DEFAULT_LINK_DIR}`, if --downloadArtifacts is passed"
    f" the value defaults to `{DEFAULT_WITH_ARTIFACTS_LINK_DIR}`",
)
@click.option(
    "-e",
    "--edition",
    type=click.Choice(EDITIONS, case_sensitive=False),
    default=DEFAULT_EDITION,
    help=f"Edition of the build to download.",
)
@click.option(
    "-p",
    "--platform",
    help=f"Platform to download. Available platforms can be found in {SETUP_REPRO_ENV_CONFIG_FILE}.",
)
@click.option(
    "-a",
    "--architecture",
    type=click.Choice(ARCHITECTURES, case_sensitive=False),
    default=DEFAULT_ARCHITECTURE,
    help="Architecture to download.",
)
@click.option(
    "-v",
    "--variant",
    help="Specify a variant to use, which supersedes the --platform, --edition and --architecture options.",
)
@click.option(
    "--installLastLTS",
    "install_last_lts",
    is_flag=True,
    help="If specified, the last LTS version will be installed.",
)
@click.option(
    "--installLastContinuous",
    "install_last_continuous",
    is_flag=True,
    help="If specified, the last continuous version will be installed.",
)
@click.option(
    "-sb",
    "--skipBinaries",
    "skip_binaries",
    is_flag=True,
    help="Whether to skip downloading binaries.",
)
@click.option(
    "-ds",
    "--downloadSymbols",
    "download_symbols",
    is_flag=True,
    help="Whether to download debug symbols.",
)
@click.option(
    "-da",
    "--downloadArtifacts",
    "download_artifacts",
    is_flag=True,
    help="Whether to download artifacts.",
)
@click.option(
    "-dv",
    "--downloadPythonVenv",
    "download_python_venv",
    is_flag=True,
    help="Whether to download python venv.",
)
@click.option(
    "-ec",
    "--evergreenConfig",
    "evergreen_config",
    type=click.Path(),
    help=f"Location of evergreen configuration file.",
    show_default=f"If not specified it will look for it in the following locations: {EVERGREEN_CONFIG_LOCATIONS}.",
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    help="Set DEBUG logging level.",
)
@click.option(
    "-rp",
    "--require-push",
    is_flag=True,
    help="Require the push task to be successful for assets to be downloaded.",
)
@click.option(
    "--resmokeCmd",
    "resmoke_cmd",
    default=DEFAULT_RESMOKE_CMD,
    help="Command to invoke resmoke.py.",
)
@click.option(
    "--fallbackToMaster",
    "fallback_to_master",
    is_flag=True,
    help="Fallback to downloading the latest binaries from master if the requested"
    " version is not found (Only application for mongo versions).",
)
@click.option("--evgVersionsFile", "evg_versions_file", type=click.Path(), hidden=True)
def setup_repro_env(
    ctx: click.Context,
    install_dir: str,
    link_dir: str,
    edition: str,
    platform: Optional[str],
    architecture: str,
    variant: Optional[str],
    versions: List[str],
    install_last_lts: bool,
    install_last_continuous: bool,
    skip_binaries: bool,
    download_symbols: bool,
    download_artifacts: bool,
    download_python_venv: bool,
    evergreen_config: Optional[str],
    debug: bool,
    require_push: bool,
    resmoke_cmd: str,
    fallback_to_master: bool,
    evg_versions_file: str,
) -> None:
    """
    Setup MongoDB repro environment.

    Downloads and installs particular MongoDB versions into install directory
    and symlinks the binaries to another directory. Each symlink name will
    include the binary release version, e.g. mongod-6.1. This script supports
    community and enterprise builds.

    Accepts binary versions, `master`, full git commit hashes, evergreen
    version ids, evergreen task ids.

    Binary version examples: <major.minor>, 4.2, 4.4, 5.0 etc.

    If no version is specified the last LTS and the last continuous versions
    will be installed.
    """

    if not versions:
        install_last_lts = True
        install_last_continuous = True
    if ctx.get_parameter_source("install_dir") == ParameterSource.DEFAULT and download_artifacts:
        install_dir = DEFAULT_WITH_ARTIFACTS_INSTALL_DIR
    if ctx.get_parameter_source("link_dir") == ParameterSource.DEFAULT and download_artifacts:
        link_dir = DEFAULT_WITH_ARTIFACTS_LINK_DIR

    setup_logging(debug)

    download_options = DownloadOptions(
        download_binaries=(not skip_binaries),
        download_symbols=download_symbols,
        download_artifacts=download_artifacts,
        download_python_venv=download_python_venv,
        install_dir=os.path.abspath(install_dir),
        link_dir=os.path.abspath(link_dir),
    )
    setup_repro_params = SetupReproParameters(
        edition=edition.lower() if edition else None,
        platform=platform.lower() if platform else None,
        architecture=architecture.lower() if architecture else None,
        variant=variant.lower() if variant else None,
        versions=versions,
        install_last_lts=install_last_lts,
        install_last_continuous=install_last_continuous,
        ignore_failed_push=(not require_push),
        fallback_to_master=fallback_to_master,
        download_options=download_options,
        evg_version_file=evg_versions_file,
    )
    evg_api = get_evergreen_api(evergreen_config)

    def dependencies(binder: inject.Binder) -> None:
        """Define dependencies for execution."""
        binder.bind(SetupReproEnvConfig, SETUP_REPRO_ENV_CONFIG)
        binder.bind(EvergreenApi, evg_api)
        binder.bind(ResmokeProxy, ResmokeProxy.with_cmd(resmoke_cmd))

    inject.configure(dependencies)
    setup_repro_orchestrator = inject.instance(SetupReproOrchestrator)

    success = setup_repro_orchestrator.execute(setup_repro_params)
    if not success:
        sys.exit(1)
